/*
   ccs_Bde.cc

   =============================================================================

   Description:

   Custom class to manage the current BDE session directories enabling dBase to
   more fully comply with the Microsoft security model.

   The current BDE session is NOT the same as the current dBase session. The
   current dBase session class wraps the current BDE session adding properties
   specific to dBase and hiding current BDE session properties.

   One of the BDE session properties can be used to change the folder the BDE
   uses to create temporary objects like temporary SQL files. By default the BDE
   creates temporary objects in the current BDE session default database working
   directory or folder. This is set at application start to the application
   startup folder.

   Applications requiring Microsoft security model compliance MUST be installed
   into CSIDL_PROGRAM_FILES and are NOT permitted to write into their startup
   folder.

   dBase (via the BDE) violates this rule and placing temporary objects into the
   application startup folder may present UAC confirm prompts to the user
   if they have an admin access token or generate an access violation if not.

   Where the BDE temporary objects are created can be changed by using the BDE
   private directory, normally only applicable to Paradox tables, but still used
   by all BDE drivers.

   This class surfaces access to the BDE Paradox driver NET DIR property and the
   BDE private directory for BDE temporary objects.

   Dependencies: :dUFLP:rmAlias.cc

   =============================================================================

   History:

      Author:        Andrew Shimmin
      Date:          23-Aug-2008
      Version:       0.0.0.1

      Description:   Original release

      Author:        Andrew Shimmin
      Date:          24-Aug-2008
      Version:       0.0.0.2

      Description:   Added comments

      Author:        Andrew Shimmin
      Date:          26-Aug-2008
      Version:       0.0.0.3

      Description:   Added active data object checking and precedence compliance
                     for SetSessionNetworkDirectory() & SetSessionPrivateDirectory()

      Author:        Andrew Shimmin
      Date:          26-Nov-2008
      Version:       0.0.0.4

      Description:   Added missing DbiGetErrorString() external

      Author:        Andrew Shimmin
      Date:          4-Dec-2008
      Version:       0.0.0.5

      Description:   Added additional comments on placement of the BDE network
                     and private directories.

      Author:        Andrew Shimmin
      Date:          15-Mar-2009
      Version:       0.0.0.6

      Description:   Added Bde class property lBugMessages to enable suppression
                     of BUG messages.

      Author:        Andrew Shimmin
      Date:          20-Mar-2009
      Version:       0.0.0.7

      Description:   Moved BdeSession::SetPrivateDirectory() comments to the Bde
                     class commentary.

      Author:        Andrew Shimmin
      Date:          21-Mar-2009
      Version:       0.0.0.8

      Description:   Added class methods:

                     class::IsPrivateDirectorySet()

                     Returns TRUE/FALSE for session private directory set/reset.

                     class::IsAdministratorAccount()

                     Returns TRUE for caller is logged on to the Administrator
                     account.

                     Added check for a windows Administrator logged on account
                     in the methods:

                     class::SetSessionNetworkDirectory()
                     class::SetSessionPrivateDirectory()

                     See comments below.

      Author:        Andrew Shimmin
      Date:          28-Sep-2010
      Version:       0.0.0.9

      Description:   Added methods getStatus() and getErrorMessage().

                     Subclassed rmAlias_v1.07.0520a.cc from the timer class with:

                     class rmAlias of Timer() custom

                     and set its version to 1.07.0520a

      Author:        Andrew Shimmin
      Date:          28-Sep-2010
      Version:       0.0.0.10

      Description:   Renamed the class to ccs_Bde and its source file to ccs_Bde.cc

      Author:        Andrew Shimmin
      Date:          29-Sep-2010
      Version:       0.0.0.11

      Description:   Added instanceId support.
                     Added messaging methods.

      Author:        Andrew Shimmin
      Date:          16-Oct-2010
      Version:       0.0.0.12

      Description:   Added BDE unresolvable issues comments on IDE program
                     development and testing when _QSQL*.DBT files are produced.

                     Added CheckTemporaryMemoFiles() method.

      Author:        Andrew Shimmin
      Date:          3-Nov-2010
      Version:       0.0.0.14

      Description:   Added properties:

                     nBdeApiError
                     cBdeApiError

                     Added methods:

                     getBdeApiError()
                     getBdeApiErrorMessage()
                     resetBdeApiError()

                     to provide access to the class Bde API status returned.

                     Added methods:

                     createAlias()
                     createDbaseAlias()
                     deleteAlias()

                     to provide Bde API status returned.

      Author:        Andrew Shimmin
      Date:          28-Dec-2010
      Version:       0.0.0.15

      Description:   Added check for default dBPLUS session only in methods:

                     setSessionNetworkDirectory()
                     setSessionPrivateDirectory()

      Author:        Andrew Shimmin
      Date:          10-Jan-2011
      Version:       0.0.0.16

      Description:   Added BdeConfiguration class.
                     Added getConfigurationSetting() method.
                     Added getConfigurationDbaseTableLevel() method.

      Author:        Andrew Shimmin
      Date:          4-Mar-2011
      Version:       0.0.0.17

      Description:   Added setCurrentSession() method.

      Author:        Andrew Shimmin
      Date:          19-Mar-2011
      Version:       0.0.0.18

      Description:   Added createDbaseAliasPersistent() method.

      Author:        Andrew Shimmin
      Date:          26-Mar-2011
      Version:       0.0.0.19

      Description:   Added fixDbiDeleteAliasBug() method.
                     Modified restartBDE() method to support fixDbiDeleteAliasBug() method.
                     Modified class initialisation to support dBPLUS v2.70 UAC support.

                     Note: Comments have not been updated for this version.

      Author:        Andrew Shimmin
      Date:          7-Apr-2011
      Version:       0.0.0.20

      Description:   Added showClientInformation() method.

                     Added cClientConfigFileFullPath property for default BDE
                     configuration file location used in the restartBDE() method.
                     Changed DBIINTFVER to 500.
                     Added getSystemConfigurationInformation() method.
                     Added showSystemConfigurationInformation() method.

                     Note: Comments have not been updated for this version.

      Author:        Andrew Shimmin
      Date:          17-Apr-2011
      Version:       0.0.0.21

      Description:   Removed cClientConfigFileFullPath property for default BDE
                     configuration file location.
                     Modified BDE configuration file location to the cSystemIniFile
                     property, returned from getSystemConfigurationInformation()
                     method called by the initialiseObject() method.

                     Added getSesCfgMode2() and showSesCfgMode2() methods.
                     Modified deleteAlias() method to fix the dBPLUS v2.70 bug

                     Removed createDbaseAliasPersistent() method so ONLY persistent
                     BDE aliases can be created. Non-persistent aliases will need
                     to use the session.AddAlias() and session.DeleteAlias() methods.

                     Note: Comments have not been updated for this version.

      Author:        Andrew Shimmin
      Date:          22-Apr-2011
      Version:       0.0.0.22

      Description:   Removed fixDbiDeleteAliasBug() call from initialiseObject()
                     and close() methods.

                     Added getCurrentSessionHandle() and IsDbaseUACSupported() methods.
                     Renamed BitTest() method to IsBitSet().
                     Re-wrote deleteAlias() method.

                     Added ModifyConfigurationSetting() method for Rich so he
                     could explore BDE settings management from within dBPLUS.

                     e.g.

                     local cNodeValue

                     cNodeValue = []

                     b = new ccs_Bde()

                     ?b.getConfigurationSetting([\system\init],[maxfilehandles],cNodeValue,[display])

                     ?cNodeValue
                     cNodeValue = [2048]

                     ?b.modifyConfigurationSetting([\system\init],[maxfilehandles],cNodeValue,[display])
                     ?b.getConfigurationSetting([\system\init],[maxfilehandles],cNodeValue,[display])

                     b.close()

                     Note: Comments have not been updated for this version.

   =============================================================================

   Restrictions:

      This class is only supported on Windows 2000 and above.

      This class does NOT support multiple dBPLUS sessions objects. ONLY the default
      dBPLUS session object is supported. dBPLUS v2.62 does NOT surface the BDE
      API dbiSetCurrentSession() so to enable multiple session object support is
      convoluted. Maybe in a subsequent version of ccs_bde.cc it can be provided
      if it becomes critical to do so.

      Setting the BDE private directory will create the PDOXUSRS.NET network
      lock file and can be placed in CSIDL_LOCAL_APPDATA.

      The private directory MUST be available for exclusive access. NO OTHER
      BDE users can access the private directory and therfore MUST placed in
      CSIDL_LOCAL_APPDATA.

      Placement of these directories is important. I recommend they be placed in
      a subfolder of the CSIDL_LOCAL_APPDATA and I use the following location:

      CSIDL_LOCAL_APPDATA+[\]+cCompanyName+[\]+cApplicationName+[\dBase\BdeNetworkDirectory]
      CSIDL_LOCAL_APPDATA+[\]+cCompanyName+[\]+cApplicationName+[\dBase\BdePrivateDirectory]

      Testing showed keeping the BDE network and private directories separated is
      best practise.

      All database objects MUST be INACTIVE (d.active := false) BEFORE the method
      SetSessionPrivateDirectory("TheFolderYouWant") is called to set the BDE
      private directory and again before SetSessionPrivateDirectory() is called
      to reset the BDE private directory to its default. Class methods check for
      any active database and query objects and return an error if any are
      found.

      Failure to manage these error returns from SetSessionNetworkDirectory() and
      SetSessionPrivateDirectory() MAY render the BDE and dBASE unstable and
      produce unpredictable results.

   =============================================================================

   IMPORTANT:  BDE unresolvable issues

   =============================================================================

   BDE temporay _QSQL*.DBT files

      If any BDE temporary _QSQL*.DBT files are generated, normally because you
      have memo fields in a table and an ORDER BY clause is used to SELECT a
      query record set, The BdePrivateDirectory cannot be released until the
      application exits as the BDE creates a persistent setting of some sort if
      the program is run with elevated privilege.

      This is not a problem for an executable, however, it poses a problem when
      testing in the IDE.

      After resetting the BdePrivateDirectory when using the IDE it is CRITICAL
      to quit the IDE on program exit to release the BdePrivateDirectory

      This can be done with the following:

         if ([runtime]$lower(version(0))) = 0         // IDE?
            QUIT
         endif

         RETURN

      It is highly recommended to ALWAYS exit the IDE with a QUIT command when
      testing with the BdePrivateDirectory.

      Failure to comply with the above will prevent the BdePrivateDirectory from
      being set on the next program run.

   =============================================================================

   Logged into the Administrator account:

      When logged into the windows Administrator account the
      class::SetSessionPrivateDirectory() method throws the error:

      "Directory is controlled by other .NET file."

      This error does not occur when logged into other accounts with or without
      administrative privilege.

      When the error occurs it leaves a BDE corruption which can only be resolved
      by reinstallation of the BDE.

      To PREVENT the error occuring the class will:

         On instantiation, set the class properties:

            cStatus        = [error]
            cErrorMessage  = [Administrator account logon NOT supported]

         if the caller is logged on to the Administrator account.

         the class methods SetSessionNetworkDirectory() & SetSessionPrivateDirectory()
         will return FALSE and set the class properties:

            cStatus        = [error]
            cErrorMessage  = [Administrator account logon NOT supported]

         if the caller is logged on to the Administrator account.

      IT IS NOT POSSIBLE TO SET EITHER THE NETWORK OR PRIVATE DIRECTORIES WHEN
      LOGGED IN TO THE ADMINISTRATOR ACCOUNT.

   =============================================================================

   Accessing tables in the startup folder:

      dBase application initialisation sets the BDE private and working
      directories to the application startup folder. dBase also sets the
      default database (_app.databases[1].handle) working directory
      to the application startup folder. This enables the dBase navigator
      and applications to access tables, without using a BDE alias, in the
      current directory.

      DbiSetPrivateDir() sets the working directory for the default
      database to the same folder as the private directory. This change
      will PREVENT the dBase navigator and applications from accessing tables
      without using a BDE alias.

      A DbiSetPrivateDir(null) reset sets the default database working directory
      to the default startup directory. For PLUS.exe this is the PLUS.exe
      startup folder "C:\Program Files\dBase\BIN" and for PLUSrun.exe it is
      where ever PLUSrun.exe started.

      This can produce unpredictable results in the IDE and runtime.

      To alleviate this problem the default database working directory is set
      to set("directory") when DbiSetPrivateDir(null) is used to reset the BDE
      private directory. This strategy returns the BDE working directory to
      the same as the dBase application initialisation.

      There are some subtle implications to setting the BDE private directory.

      When DbiSetPrivateDir() sets the private directory it sets the default
      database working directory to the same as the private directory.

      ANY FURTHER ACCESS TO TABLES IN THE APPLICATION STARTUP FOLDER IS NOT
      POSSIBLE UNLESS A BDE ALIAS OR AN ABSOLUTE PATH IS USED TO ACCESS THEM.

   =============================================================================

   Acknowlegements:

      This class is a derived class from rmAlias.cc.

      The latest version of rmAlias.cc is available here:

      http://www.dbasedeveloper.co.uk/rmAlias/rmAlias.htm

   =============================================================================

   Disclaimer:

      The author of this custom class will NOT be held responsible or accountable
      for any effect or impact of its use and any risk resulting from its
      application is entirely the responsibility of the user.

      If you do not agree then do NOT use it.

   =============================================================================

   Usage:

      local oBde,cError,lPrivateDirectoryDBTFiles

      if type([ccs_Bde::IamLoaded]) # [FP]
         set procedure to :YourSourceAlias:ccs_Bde.cc additive
      endif

      oBde = new ccs_Bde()                                                    // instantiate a ccs_Bde object

      if oBde.cStatus == [ok]                                                 // instantiate correctly?

         if oBde.SetSessionNetworkDirectory([TheFolderYouWantForIt])          // set the BDE NetworkDirectory? (paradox driver NET DIR property)
            if .not.oBde.SetSessionPrivateDirectory([TheFolderYouWantForIt])  // set the BDE PrivateDirectory?
               cError = oBde.getErrorMessage()                                // get error message

               ... manage error ...

            endif
         else
            cError = oBde.getErrorMessage()                                   // get error message

            ... manage error ...

         endif

         if oBde.IsPrivateDirectorySet                                        // PrivateDirectory set ok?

            ... set your database objects active ...
            ... set your query objects active ...

            ... process your application ...

            ... set your query objects inactive ...
            ... set your database objects inactive ...

            oBde.SetSessionPrivateDirectory()                                 // reset the BDE PrivateDirectory

         endif

      else
         ?[ccs_Bde.cc instantiation error: ]+oBde.getErrorMessage()
      endif

      lPrivateDirectoryDBTFiles = oBde.CheckTemporaryMemoFiles()              // remember any temporary memo files

      oBde.close()
      oBde := null

      // on program exit

      if lPrivateDirectoryDBTFiles                                            // any temporary memo files?
         if ([runtime]$lower(version(0))) = 0                                 // ide?
            quit
         else
            return
         endif
      endif

      return


      There is an example of its application in:

      TestVista.prg and TestBdeClass.prg

   =============================================================================

*/

   // messaging constants

   #define  M_MB_OK                    0x00000000
   #define  M_MB_ICONHAND              0x00000010                             // Stop-sign icon.
   #define  M_MB_ICONQUESTION          0x00000020                             // Question-mark icon.
   #define  M_MB_ICONEXCLAMATION       0x00000030
   #define  M_MB_ICONINFORMATION       0x00000040

   #define  M_MB_ABORTRETRYIGNORE      0x00000002                             // three push buttons: Abort, Retry, and Ignore.
   #define  M_MB_YESNO                 0x00000004                             // two push buttons: Yes and No.
   #define  M_MB_RETRYCANCEL           0x00000005                             // two push buttons: Retry and Cancel.

   #define  M_MB_TASKMODAL             0x00002000
   #define  M_MB_TOPMOST               0x00040000
   #define  M_MB_SETFOREGROUND         0x00010000

   #define  M_IDOK                     1                                      // OK button was selected.
   #define  M_IDCANCEL                 2                                      // Cancel button was selected.
   #define  M_IDABORT                  3                                      // Abort button was selected.
   #define  M_IDRETRY                  4                                      // Retry button was selected.
   #define  M_IDIGNORE                 5                                      // Ignore button was selected.
   #define  M_IDYES                    6                                      // Yes button was selected.
   #define  M_IDNO                     7                                      // No button was selected.
   #define  M_IDCLOSE                  8                                      // Close button was selected.
   #define  M_IDHELP                   9                                      // Help button was selected.

   // display constants

   #define DISPLAY_BDE_COLUMN2            65

   // BDE environmental

   #define  DBIINTFVER                 500                                    // Version of the interface

   // BDE errors

   #define DBIERR_NONE                 0
   #define DBIERR_INVALIDHNDL          9990
   #define DBIERR_EOF                  8706
   #define DBIERR_OBJNOTFOUND          8712

   // BDE constants

   #define DBIMAXNAMELEN               31                                     // Name limit (table, field etc)
   #define DBIMAXSPNAMELEN             64                                     // maximum stored procedure name length
   #define DBIMAXFUNCNAMELEN           255                                    // maximum function name length
   #define DBIMAXFLDSINKEY             16                                     // maximum fields in a key
   #define DBIMAXKEYEXPLEN             220                                    // maximum Key expression length
   #define DBIMAXEXTLEN                3                                      // maximum file extension length, not incl. dot (excluding zero termination)

   #define DBIMAXTBLNAMELEN            260                                    // maximum table name length
   #define DBIMAXDRIVELEN              127                                    // maximum drive length
   #define DBIMAXPATHLEN               260                                    // maximum path+file name length (excluding zero termination)

   #define DBIMAXMSGLEN                127                                    // maximum message length
   #define DBIMAXVCHKLEN               255                                    // maximum val check length
   #define DBIMAXPICTLEN               175                                    // maximum picture length
   #define DBIMAXFLDSINSEC             256                                    // maximum fields in security spec

   #define DBIMAXSCFIELDS              32                                     // maximum nbr of fields in a config section
   #define DBIMAXSCFLDLEN              128                                    // maximum field length
   #define DBIMAXSCRECSIZE             DBIMAXSCFIELDS*DBIMAXSCFLDLEN          // maximum rec size

   #define DBIMAXUSERNAMELEN           14                                     // maximum user name (general)
   #define DBIMAXXBUSERNAMELEN         16                                     // maximum user name length (xBASE)
   #define DBIMAXBOOKMARKLEN           4104                                   // maximum Bookmark length

   #define DBIMAXTRACELEN              8192                                   // maximum trace msg length

   #define DBIMAXTYPEDESC              127                                    // maximum Type Description size
   #define DBIMAXDRSQLSTR              8192                                   // maximum Size of SQL Constraint

   #define DBIMAXOLEDBNAMELEN          64                                     // maximum OLE DB name length

   // BDE session properties

   #define  sesMAXPROPS                0x00020000                             // ro UINT16, Number of defined properties
   #define  sesSESSIONNAME             0x00020001                             // ro pDBINAME , Name of session
   #define  sesNETFILE                 0x00020002                             // rw pCHAR, Full filename of netfile (MAXPATHLEN)
   #define  sesCFGMODE                 0x00020003                             // rw CFGMode, Mode of configuration file.
   #define  sesDEFAULTDRIVER           0x00020004                             // rw pDBINAME, default driver name
   #define  sesGROUPNAME               0x00020005                             // rw pCHAR, Security - user group name (dBASE)
   #define  sesUSERNAME                0x00020006                             // rw pCHAR, User Name
   #define  sesUSERPASSWORD            0x00020007                             // rw pCHAR, User password
   #define  sesSECFILENAME             0x00020008                             // rw pCHAR, Location of dBASE Security file
   #define  sesDRNAME                  0x00020009                             // rw pCHAR, Repository Database Name
   #define  sesCFGUPDATE               0x0002000A                             // rw CFGUpdate, Mode for updating session configuration
   #define  sesCFGNAME                 0x0002000B                             // ro CFGName, get session configuration name
   #define  sesCFGMODE2                0x0002000C                             // rw new CFGMode (persist, session, virtual)
   #define  sesCFGRefresh              0x0002000F                             // rw, set session to refresh virtual config
   #define  maxsesPROPS                13                                     // keep in sync when adding session properties

   // BDE cursor properties

   #define dbiREADWRITE                0                                      // Read + Write   (Default)
   #define dbiREADONLY                 1                                      // Read only

   #define dbiNOLOCK                   0                                      // No lock   (Default)
   #define dbiWRITELOCK                1                                      // Write lock
   #define dbiREADLOCK                 2                                      // Read lock

   // BDE configuration CFGMode info list constants

   #define cfgPersistent               0                                      // Persistent only
   #define cfgSession                  1                                      // Session relative only
   #define cfgAll                      2                                      // All (union)

   // BDE configuration CFGMode2 info list constants

   #define cfgmNone                    0x00
   #define cfgmVirtual                 0x01
   #define cfgmPersistent              0x02
   #define cfgmSession                 0x04
   #define cfgmAll                     cfgmVirtual+cfgmPersistent+cfgmSession

   // BDE buffer length definitions

   #define DBIPATH                     DBIMAXPATHLEN+1                        // DOS path
   #define DBINAME                     DBIMAXNAMELEN+1                        // name
   #define DBIEXT                      DBIMAXEXTLEN+1                         // file extension
   #define DBIDOTEXT                   DBIMAXEXTLEN+2                         // file extension including "."
   #define DBIDRIVE                    DBIMAXDRIVELEN+1                       // drive name
   #define DBITBLNAME                  DBIMAXTBLNAMELEN+1                     // table name
   #define DBIUSERNAME                 DBIMAXUSERNAMELEN+1                    // user name
   #define DBIKEY                      DBIMAXFLDSINKEY                        // list of fields in a key
   #define DBIKEYEXP                   DBIMAXKEYEXPLEN+1                      // key expression

   #define DBIVCHK                     DBIMAXVCHKLEN+1                        // validity check
   #define DBIPICT                     DBIMAXPICTLEN+1                        // picture clause
   #define DBIMSG                      DBIMAXMSGLEN+1                         // error message

   // BDE configuration strings

   #define cDbaseTableCreate           [\DRIVERS\DBASE\TABLE CREATE]

class ccs_Bde(vInstanceId) of rmAlias from :dUFLP:rmAlias.cc                         // use rmAlias.cc derived from the timer class

/*
   =============================================================================

   Class constructor

   =============================================================================
*/

   // custom properties

*  this.className = [ccs_Bde]                                                 // used ONLY when derived from rmAlias.cc

   this.cVersion = [0.0.0.22]
   this.cVersionDate = [19-Apr-2011]

   this.nMethods = 0                                                          // class method count

   this.cStatus = [ok]                                                        // class status
   this.cErrorMessage = []                                                    // class error text

   this.nBdeApiError = 0                                                      // checkDbi() Bde API status results
   this.cBdeApiError = []

   this.cClientName = []                                                      // BDE client properties
   this.nClientSessions = 0
   this.cClientWorkingDirectory = []
   this.cClientLanguage = []

   this.lClientEnvForceLocalInit = true                                       // BDE client initialisation setting

   this.lSystemLocalShare = false                                             // BDE system configuration properties
   this.nSystemNetProtocol = 0
   this.lSystemNetShare = false
   this.cSystemNetType = []
   this.cSystemUserName = []                                                  // network user name
   this.cSystemIniFile = []                                                   // system configuration file full path
   this.cSystemLanguage = []

   this.cSessionNetworkDirectorySaved = []                                    // saved session network directory
   this.cSessionPrivateDirectorySaved = []                                    // saved session private directory
   this.cSessionNetworkDirectory = []                                         // current session network directory
   this.cSessionPrivateDirectory = []                                         // current session private directory

   this.hMessageParentWindow = 0
   this.lAdministratorAccount = false

   this.lBugMessages = true                                                   // issue bug messages
   this.lDebugger = true                                                      // start debugger

   this.lActiveDataObjects = false                                            // remember active data object found
   this.cTemporaryObjectsFolder = []                                          // remember private directory
   this.lTemporaryObjectsFolderSet = false                                    // remember private directory state
   this.lTemporaryMemoFiles = false                                           // remember private directory has _QSQL*.DBT files
   this.lTemporaryObjectsFolderDelete = true                                  // remember private directory folder delete on reset

   this.nDbaseTableLevel = 0                                                  // configuration file dBase driver table level
   this.cDbaseLanguageDriver = []                                             // configuration file dBase driver language driver

   // The following property was added on 26-Mar-2011 during testing to
   // identify a fix forthe dBPLUS v2.70 (b2160) DbiDeleteAlias() bug.
   //
   // The DbiDeleteAlias() bug was found on 15-Apr-2011 and it is no longer
   // needed so ... do NOT set it to TRUE.

   this.lFixDbIDeleteAliasBugEnabled = false                                  // call the FixDbIDeleteAliasBug() method on initialise/close

   /*

   The following property provides the ability to locate this class instance using:

      local lFound,oInstance

      lFound = false
      oInstance = findInstance([ccs_Bde])

      do while (.not.empty(oInstance)).and.(.not.lFound)
         if oInstance.instanceId == [YourUniqueInstanceIdentifier] - note case sensitive
            lFound = true
         else
            oInstance = findInstance([ccs_Bde],oInstance)
         endif
      enddo

   */

   this.instanceId = [ccs_Bde]                                                // default property for findInstance() searching
   this.lEnableInstanceIdWarning = false                                      // set to false to disable warning messages

   // class protection

*   protect  GetDatabaseFolders,GetSessionDatabases,GetSessionHandle,GetSessionId,;
            GetSessionNetworkDirectory,GetSessionPrivayeDirectory,TemporaryObjectsFolderDelete,;
            CheckAdministratorAccount

*   protect  this.lAdministratorAccount,;
            this.lActiveDataObjects,this.lTemporaryObjectsFolder,lTemporaryObjectsFolderSet,;
            this.lTemporaryMemoFiles

   // externals

   if type([DbiGetErrorString]) # [FP]
      extern cushort DbiGetErrorString(cword,cstring) idapi32
   endif
   if type([DbiGetClientInfo]) # [FP]
      extern cushort DbiGetClientInfo(cptr) idapi32
   endif
   if type([DbiGetSysConfig]) # [FP]
      extern cushort DbiGetSysConfig(cptr) idapi32
   endif
   if type([DbiGetSesInfo]) # [FP]
      extern cushort DbiGetSesInfo(cptr) idapi32
   endif
   if type([DbiGetCurrSession]) # [FP]
      extern cushort DbiGetCurrSession(cptr cLong) idapi32
   endif
   if type([DbiSetCurrSession]) # [FP]
      extern cushort DbiSetCurrSession(cLong) idapi32
   endif
   if type([DbiSetPrivateDir]) # [FP]
      extern cushort DbiSetPrivateDir(cstring) idapi32
   endif
   if type([DbiGetCfgMode2Prop]) # [FP]
      extern cushort DbiGetCfgMode2Prop(chandle,culong,cptr culong,cushort,cptr cushort) idapi32 from [DbiGetProp]
   endif
   if type([DbiSetNetDirProp]) # [FP]
      extern cushort DbiSetNetDirProp(chandle,clong,cstring) idapi32 from [DbiSetProp]
   endif
   if type([DbiSetProp]) # [FP]
      extern cushort DbiSetProp(chandle,clong,clong) idapi32 from [DbiSetProp]
   endif
   if type([DbiGetDirectory]) # [FP]
      extern cushort DbiGetDirectory(chandle,clogical,cstring) idapi32
   endif
   if type([DbiSetDirectory]) # [FP]
      extern cushort DbiSetDirectory(chandle,cstring) idapi32
   endif

   if type([DbiOpenCfgInfoList]) # [FP]
      extern cushort DbiOpenCfgInfoList(chandle,cshort,cshort,cstring,cptr) idapi32
   endif
   if type([DbiSetToBegin]) # [FP]
      extern cushort DbiSetToBegin(chandle) idapi32
   endif
   if type([DbiGetNextRecord]) # [FP]
      extern cushort DbiGetNextRecord(chandle,cshort,cptr,cptr) idapi32
   endif
   if type([DbiModifyRecord]) # [FP]
      extern cushort DbiModifyRecord(chandle,cptr,clogical) idapi32
   endif
   if type([DbiCloseCursor]) # [FP]
      extern cushort DbiCloseCursor(cptr) idapi32
   endif

   if type([RemoveDirectoryA]) # [FP]
      extern cLogical RemoveDirectoryA(cString) kernel32
   endif

   // initialise class

   class::fileMethodCount()
   class::initialiseObject(vInstanceId)

/*
   =============================================================================

   Class methods

   =============================================================================
*/

// Method ======================================================================

   function IamLoaded()
      return

// Method ======================================================================

   function fileMethodCount()

      local f,l

      f = []
      l = []

      if ([runtime]$lower(version()) = 0).and.(lower(right(program(0),2)) == [cc])

         f := new file()
         f.open(program(0),[R])

         try
            do while .not.f.eof()
               l := f.gets()
               if (left(l,2) == "//").and.(lower(substr(l,4,6)) == [method])
                  this.nMethods++
               endif
            enddo
         catch (exception e)
         finally
            f.close()
         endtry

      endif

      return

// Method ======================================================================

   function IsBitSet(nMask,nValue)
   /*

      Method to perform a bit test.

      Input:

         nMask    =  bit mask to use
         nValue   =  value to test if the nMask bits are set

      Output:

         true     =  the nMask bits are set in nValue
         false    =  the nMask bits are NOT set in nValue

   */

      return (bitAnd(nMask,nValue) = nMask)

// Method ======================================================================

   function getVersion()
      return (this.cVersion)

// Method ======================================================================

   function getVersionDate()
      return (this.cVersionDate)

// Method ======================================================================

   function getStatus()
      return (this.cStatus)

// Method ======================================================================

   function getErrorMessage()
      return (this.cErrorMessage)

// Method ======================================================================

   function resetStatus()

      this.cStatus := [ok]
      this.cErrorMessage := []

      return

// Method ======================================================================

   function setStatus(cStatus,cErrorMessage)

      this.cStatus := iif(.not.empty(cStatus),trim(cStatus),[ok])
      this.cErrorMessage := iif(.not.empty(cErrorMessage),trim(cErrorMessage),[])

      return

// Method ======================================================================

   function getBdeApiError()
      return (this.nBdeApiError)

// Method ======================================================================

   function getBdeApiErrorMessage()
      return (this.cBdeApiError)

// Method ======================================================================

   function resetBdeApiError()

      this.nBdeApiError := DBIERR_NONE
      this.cBdeApiError := []

      return

// Method ======================================================================

   function showError(cFunction)

      if type([_app.eAbortException]) == [O].and._app.eAbortException.baseClassName == [EXCEPTION]
         _app.eAbortException.message := this.cErrorMessage
         _app.eAbortException.cFunction := cFunction+[()]
         throw _app.eAbortException                                           // abort
      else
         if this.lBugMessages                                                 // issue bug message?
            msgbox(this.cErrorMessage+[.],cFunction+[()],16)
            if ([runtime]$lower(version(0))) = 0
               if this.lDebugger
                  set step on
               endif
               ?program()+[ - BUG - ]+this.cErrorMessage
               ?
            endif
         endif
      endif

      this.cStatus := iif(.not.empty(this.cErrorMessage),[error],[ok])        // return class status

      return

// Method ======================================================================

   function IsAdministratorAccount()
      return (class::CheckAdministratorAccount())

// Method ======================================================================

   function checkAdministratorAccount()

      this.lAdministratorAccount := (lower(getenv([username])) == [administrator])

      if this.lAdministratorAccount
         this.cStatus := [error]
         this.cErrorMessage := [Administrator account logon NOT supported]
      else
         this.cStatus := [ok]
         this.cErrorMessage := []
      endif

      return (this.lAdministratorAccount)

// Method ======================================================================

   function initialiseObject(vInstanceId)

      local lReturn,cNodeValue

      lReturn = true
      cNodeValue = []

      if type([MessageBox]) # [FP]
         extern cint MessageBox(chandle,cstring,cstring,cint) user32 from [MessageBoxA]
      endif

      // initialise

      vInstanceId = iif(class::InstanceIdCheck(vInstanceId),vInstanceId,[ccs_Bde])

      if .not.class::CheckAdministratorAccount()                              // not Administrator account logon?
         if class::getSystemConfigurationInformation()                        // get system configuration from BDE?
            if class::getClientInformation()                                  // get client information from BDE?
               if class::getConfigurationSetting([\DRIVERS\DBASE\TABLE CREATE],[LEVEL],cNodeValue,[messages])
                  this.nDbaseTableLevel := iif(lReturn,val(cNodeValue),this.nDbaseTableLevel)
               else
                  lReturn := false
               endif
            else
               lReturn := false
            endif
         else
            lReturn := false
         endif
      else
         lReturn := false
      endif

      // finish

      if lReturn

         class::showSystemConfigurationInformation(program())                 // show system configuration from BDE
         class::showClientInformation(program())                              // show client information from BDE

      else

         class::messageError((this.cErrormessage),program())                  // display specificclass error message
         class::setStatus([error],[Class initialisation error. Class is unusable.])
         class::messageError((this.cErrormessage),program())                  // display class initialisation error message

      endif

      return (lReturn)

// Method ======================================================================

   function instanceIdCheck(vInstanceId)

      local lReturn

      lReturn = true

      do case
         case  argcount() = 0.or.empty(vInstanceId)
               if this.lEnableInstanceIdWarning

                  this.cStatus := [error]
                  this.cErrorMessage := [No ]+this.className+[ class InstanceId provided]

                  class::messageALERT([Instantiation warning. No ]+this.className+[ class InstanceId provided.],program())

                  if ([runtime]$lower(version(0))) = 0                        // ide?
                     if this.lDebugger
                        set step on
                     endif
                  endif

                  lReturn := false

               endif
         otherwise
               this.instanceId := vInstanceId
               this.instanceId := iif(type([this.instanceId]) == [C],trim(this.InstanceId),this.instanceId)
     endcase

     return (lReturn)

// Method ======================================================================

   function IsPrivateDirectorySet()
      return (this.lTemporaryObjectsFolderSet)

// Method ======================================================================

   function close()

      local lReturn

      lReturn = true

      class::showSystemConfigurationInformation(program())                    // show system configuration from BDE
      class::showClientInformation(program())                                 // show client information from BDE

      if type([this.sCFGDesc]) # [U]
         this.sCFGDesc.release()
         this.sCFGDesc := null
      endif
      if type([this.sRECProps]) # [U]
         this.sRECProps.release()
         this.sRECProps := null
      endif
      if type([this.phCursor]) # [U]
         this.phCursor.release()
         this.phCursor := null
      endif

      try
         release object this
      catch (exception e)
      endtry

      return (lReturn)

// Method ======================================================================

   function onBeforeRelease()
      return

// Method ======================================================================

   function activeDataObjects(cMessages)

   /*

      Method to check for any active database or query objects.

      Note:

      There is one exception: a dBase session default database. This
      database object is always open. This method will ignore this database
      object.

   */

      local bResult,nActiveDatabaseObjects,nActiveQueryObjects

      bResult = false                                                         // assume none

      nActiveDatabaseObjects = 0
      nActiveQueryObjects = 0

      this.cErrorMessage = []

      // check for active data objects

      this.lActiveDataObjects := false                                        // assume none

      nActiveQueryObjects = class::EnumerateHeapClassPropertyValue([query],[active],true)             // get active query count
      nActiveDatabaseObjects = class::EnumerateHeapClassPropertyValue([database],[active],true,true)  // get active database count

      do case
         case  nActiveQueryObjects <> 0.and.nActiveDatabaseObjects <> 0
               this.lActiveDataObjects := true
               this.cErrorMessage = [Active DATABASE and QUERY objects found]
         case  nActiveQueryObjects <> 0.and.nActiveDatabaseObjects = 0
               this.lActiveDataObjects := true
               this.cErrorMessage = [Active QUERY object]+iif(nActiveQueryObjects>1,[s found],[ found])
         case  nActiveQueryObjects = 0.and.nActiveDatabaseObjects <> 0
               this.cErrorMessage = [Active DATABASE object]+iif(nActiveDatabaseObjects>1,[s found],[ found])
         otherwise
               this.lActiveDataObjects := false
      endcase

      if .not.empty(this.cErrorMessage)
         if .not.empty(cMessages).or.this.lBugMessages                        // issue bug message?
            msgbox([BUG - ]+this.cErrorMessage,program(),16)
            if ([runtime]$lower(version(0))) = 0
               if this.lDebugger
                  set step on
               endif
               ?
               ?program()+[ - BUG - ]+this.cErrorMessage
               ?
            endif
         endif
         bResult = true
      endif

      return (bResult)

// Method ======================================================================

   function checkDbaseDefaultSessionDatabaseObject(oDatabase)

   /*

      Method to check for the dBase default session database object.

      Input:

         oDatabase   =  database object reference to check

      Output:

         lReturn     =  true/false for default dbase session default database object

   */

      return (oDatabase.handle = _app.databases[1].handle)

// Method ======================================================================

   function checkDbaseSessionDefaultDatabaseObject(oDatabase)

   /*

      Method to check for a dBase session default database object.

      Input:

         oDatabase   =  database object reference to check

      Output:

         lReturn     =  true for a dbase session default database object

   */

      local lReturn

      lReturn = false

      do case
         case  oDatabase.session.handle = _app.session.handle                 // dBase default session
               lReturn := (oDatabase.handle = _app.databases[1].handle)       // set default session default database object
         otherwise                                                            // dBase non-default session
               if oDatabase.driverName == [DBASE]                             // DBASE driver?
                  if empty(oDatabase.databaseName)                            // no databaseName?
                     lReturn := true                                          // set a session default database object
                  endif
               endif
      endcase

      return (lReturn)

// Method ======================================================================

   function enumerateHeapClassPropertyValue(cClassName,cClassProperty,pClassPropertyValue,lIgnoreDefaultDatabaseObjects,aObjectsFound)

   /*

      Method to enumerate, from the object heap, a class property with a
      specific value.

      Input:

         cClassName                    =  class name to enumerate
         cClassProperty                =  class property to enumerate
         pClassPropertyValue           =  class property value to enumerate
         lIgnoreDefaultDatabaseObjects =  .not.empty() to ignore a dBase default database objects
         aObjectsFound                 =  array for any objects references found

      Output:

         nResult                       =  number of object references found
         aObjectsFound                 =  array of objects references found

   */

      local    nResult,lFound

      nResult = 0                                                             // assume none found
      lFound = false                                                          // assume none found

      private  cClass,cProperty,pValue,cInstanceProperty,oInstance

      cClass = trim(cClassName)
      cProperty = trim(cClassProperty)
      pValue = pClassPropertyValue
      pValue = iif(type("pValue") == "C",trim(pValue),pValue)
      oInstance = null

      oInstance = findInstance(cClass)                                        // get first instance

      if .not.empty(oInstance)                                                // have any instances?

         do while .not.empty(oInstance)

            cInstanceProperty = "oInstance"+"."+cProperty
            if &cInstanceProperty == pValue                                   // class property value match?
               if oInstance.className == "DATABASE"
                  if argcount() >= 4
                     if .not.empty(argvector(4))                              // check for dBase session default databases?
                        do case
                           case  class::checkDbaseSessionDefaultDatabaseObject(oInstance)
                                 lFound := .not.lIgnoreDefaultDatabaseObjects
                           otherwise
                                 lFound := true
                        endcase
                     else
                        lFound := true
                     endif
                  else
                     lFound := true
                  endif
               else
                  lFound := true
               endif
            endif

            if lFound
               nResult++                                                      // count it
               if argcount() >= 5
                  aObjectsFound.add(oInstance)
               endif
            endif

            lFound := false                                                   // reset
            oInstance := findInstance(cClass,oInstance)                       // get next instance

         enddo

      endif

      return (nResult)

// Method ======================================================================

   function getDbaseBuildNumber()
      return (int(val(stuff(version(0),1,at([b],version(0)),[]))))

// Method ======================================================================

   function IsDbaseUACSupported()
      return (type([_app.useUACPaths]) # [U])

// Method ======================================================================

   function IsDbaseUACSupportEnabled()

      local lReturn

      lReturn = false

      if class::IsDbaseUACSupported()                                         // dBPLUS UAC support version?
         if type([_app.useUACPaths]) == [L]                                   // UAC support?
            lReturn := _app.useUACPaths                                       // set UAC support enabled
         endif
      endif

      return (lReturn)

// Method ======================================================================

   function getDbaseCurrentUserHomePath()

      local cCurrentUserHomePath

      cCurrentUserHomePath = []

      do case
         case  class::getDbaseBuildNumber() > 2119                            // after 2.62
               cCurrentUserHomePath := left(_app.currentUserPath,(len(_app.currentUserPath)-4))
         otherwise                                                            // before 2.70
               cCurrentUserHomePath := _dbwinhome
      endcase

      return (cCurrentUserHomePath)

// Method ======================================================================

   function getDbaseAllUsersHomePath()

      local cAllUsersHomePath

      cAllUsersHomePath = []

      do case
         case  class::getDbaseBuildNumber() > 2119                            // after 2.62
               cAllusersHomePath := left(_app.allUsersPath,(len(_app.allUsersPath)-4))
         otherwise                                                            // before 2.70
               cAllUsersHomePath := _dbwinhome
      endcase

      return (cAllUsersHomePath)

// Method ======================================================================

   function getDbaseTableLevel()
      return (this.nDbaseTableLevel)

// Method ======================================================================

   function getConfigurationSetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

   /*

      Method to get a setting from the current BDE configuration file.

      Input:

         cBdeConfigPath    =  BDE configuration path              eg. [\DRIVERS\DBASE\TABLE CREATE]
         cNodeName         =  BDE configuration path node name    eg. [LEVEL]

         cDisplayResults   =  .not.empty() to display results in the command pane

      Output:

         bResult           =  true/false for success/error
         cNodeValue        =  BDE configuration node value        eg. [7]

      Notes:

         All node values returned are strings. Boolean node values are returned
         as [TRUE]/[FALSE].

         To get all BDE configuration paths available use:

            cBdeConfigPath = [\]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

         To get each top level configuration path node names use eg.:

            cBdeConfigPath = [\Databases]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

   */

      local bResult,oConfiguration

      bResult = true
      oConfiguration = []

      oConfiguration := new BdeConfiguration()                                // create current BDE configuration file object

      bResult := oConfiguration.GetSetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

      oConfiguration.Release()
      oConfiguration := null

      return (bResult)


// Method ======================================================================

   function modifyConfigurationSetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

   /*

      Method to modify a setting from the current BDE configuration file.

      Input:

         cBdeConfigPath    =  BDE configuration path              eg. [\SYSTEM\INIT]
         cNodeName         =  BDE configuration path node name    eg. [MAXFILEHANDLES]
         cNodeValue        =  BDE configuration node value        eg. [1024]

         cDisplayResults   =  .not.empty() to display results in the command pane

      Output:

         bResult           =  true/false for success/error

      Notes:

         End nodes are called leaf nodes. ONLY leaf nodes can be modified.

         All node values returned are strings. Boolean node values are returned
         as [TRUE]/[FALSE].

         To get all BDE configuration paths available use:

            cBdeConfigPath = [\]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

         To get each top level configuration path node names use eg.:

            cBdeConfigPath = [\Databases]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

   */

      local bResult,oConfiguration

      bResult = true
      oConfiguration = []

      oConfiguration := new BdeConfiguration()                                // create current BDE configuration file object

      bResult := oConfiguration.ModifySetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

      oConfiguration.Release()
      oConfiguration := null

      return (bResult)


// Method ======================================================================

   function GetCurrentSessionHandle(hSession)

      hsession := 0

      this.nBdeApiError := DbiGetCurrSession(hSession)

      if .not.(this.nBdeApiError == DBIERR_NONE)                              // error?
         class::CheckDbi((this.nBdeApiError),[DbiGetCurrSession])             // save it
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function getClientInformation(cDisplayResults)

   /*

      Method to get the BDE client information.

   */

      local pClientInfo

      class::resetBdeApiError()

      pClientInfo = new BdeClientInfoStructure()

      this.nBdeApiError := DbiGetClientInfo(pClientInfo)                      // get client information

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?

         this.cClientName := trim(pClientInfo.szName())
         this.nClientSessions := pClientInfo.iSessions()
         this.cClientWorkingDirectory := trim(pClientInfo.szWorkDir())
         this.cClientLanguage := trim(pClientInfo.szLang())

         if (.not.empty(cDisplayResults)).and.(([runtime]$lower(version(0))) = 0)
            class::showClientInformation(program())
         endif

      else
         class::CheckDbi((this.nBdeApiError),[DbiGetClientInfo])             //  save it
      endif

      pClientInfo.release()

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function showClientInformation(cProgram)

   /*

      Method to show the BDE client information and current session directories.

   */

      local pClientInfo,cDefault,cWorking

      class::resetBdeApiError()

      pClientInfo = new BdeClientInfoStructure()

      cDefault = []
      cWorking = []

      this.nBdeApiError := DbiGetClientInfo(pClientInfo)                      // get client information

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?
         if ([runtime]$lower(version(0)) = 0)

            ?
            ?[BDE Client Information:],iif(.not.empty(cProgram),cProgram,[]) at DISPLAY_BDE_COLUMN2
            ?
            ?[ClientName:],trim(pClientInfo.szName()) at DISPLAY_BDE_COLUMN2
            ?[ClientSessions:],ltrim(str(pClientInfo.iSessions())) at DISPLAY_BDE_COLUMN2
            ?[ClientWorkingDirectory:],trim(pClientInfo.szWorkDir()) at DISPLAY_BDE_COLUMN2
            ?[ClientLanguage:],trim(pClientInfo.szLang()) at DISPLAY_BDE_COLUMN2
            ?
            ?[CurrentSession:],ltrim(str(class::getSessionId())) at DISPLAY_BDE_COLUMN2
            ?[SessionNetworkDirectory:],class::getSessionNetworkDirectory() at DISPLAY_BDE_COLUMN2
            ?[SessionPrivateDirectory:],class::getSessionPrivateDirectory() at DISPLAY_BDE_COLUMN2

            do case
               case  class::getSessionId() = 1                                // default session
                     class::getDatabaseFolders(_app.databases[1].handle,cDefault,cWorking)
                     ?[SessionDefaultDatabase_DefaultDirectory:],trim(cDefault) at DISPLAY_BDE_COLUMN2
                     ?[SessionDefaultDatabase_WorkingDirectory:],trim(cWorking) at DISPLAY_BDE_COLUMN2
               otherwise
            endcase

            ?

         endif
      else
         class::CheckDbi((this.nBdeApiError),[DbiGetClientInfo])              //  save it
      endif

      pClientInfo.release()

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function getSystemConfigurationInformation(cDisplayResults)

   /*

      Method to get the BDE system configuration information.

   */

      local pSYSConfig

      class::resetBdeApiError()

      pSYSConfig = new BdeSYSConfigStructure()

      this.nBdeApiError := DbiGetSysConfig(pSYSConfig)                        // get system configuration information

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?

         this.lSystemLocalShare := pSYSConfig.bLocalShare()
         this.nSystemNetProtocol := [NotUsed]                                 // not used by DbiGetSysConfig()
         this.lSystemNetShare := false                                        // not used by DbiGetSysConfig()
         this.cSystemNetType := trim(pSYSConfig.szNetType())
         this.cSystemUserName := trim(pSYSConfig.szUserName())
         this.cSystemIniFile := trim(pSYSConfig.szIniFile())
         this.cSystemLanguage := trim(pSYSConfig.szLangDriver())

         if (.not.empty(cDisplayResults)).and.(([runtime]$lower(version(0))) = 0)
            class::showSystemConfigurationInformation(program())
         endif

      else
         class::CheckDbi((this.nBdeApiError),[DbiGetSysConfig])               //  save it
      endif

      pSYSConfig.release()

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function showSystemConfigurationInformation(cProgram)

   /*

      Method to show the BDE system configuration information.

   */

      local pSYSConfig

      class::resetBdeApiError()

      pSYSConfig = new BdeSYSConfigStructure()

      this.nBdeApiError := DbiGetSysConfig(pSYSConfig)                        // get client system configuration information

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?
         if ([runtime]$lower(version(0)) = 0)
            ?
            ?[BDE System Configuration Information:],iif(.not.empty(cProgram),cProgram,[]) at DISPLAY_BDE_COLUMN2
            ?
            ?[SystemLocalShare:],iif(pSYSConfig.bLocalShare(),[TRUE],[FALSE]) at DISPLAY_BDE_COLUMN2
            ?[SystemNetType:],ltrim(pSYSConfig.szNetType()) at DISPLAY_BDE_COLUMN2
            ?[SystemNetworkUserName:],trim(pSYSConfig.szUserName()) at DISPLAY_BDE_COLUMN2
            ?[SystemConfigurationFile:],trim(pSYSConfig.szIniFile()) at DISPLAY_BDE_COLUMN2
            ?
         endif
      else
         class::CheckDbi((this.nBdeApiError),[DbiGetSysConfig])               //  save it
      endif

      pSYSConfig.release()

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function getSesCfgMode2(nSesCfgMode2)

   /*

      Method to get the current session sesCFGMODE2 property.

   */

      local hSession,nCfgMode2

      hSession = 0
      nCfgMode2 = 0

      class::resetBdeApiError()

      hSession := class::getSessionHandle()

      this.nBdeApiError := DbiGetCfgMode2Prop(hSession,sesCFGMODE2,nCfgMode2,null,null)

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?
         nSesCfgMode2 := nCfgMode2                                            // return property
      else
         class::CheckDbi((this.nBdeApiError),[DbiGetCfgMode2Prop])
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function showSesCfgMode2(cProgram)

   /*

      Method to get and show the current session sesCFGMODE2 property.

   */

      local hSession,nCfgMode2

      hSession = 0
      nCfgMode2 = 0

      class::resetBdeApiError()

      hSession := class::getSessionHandle()

      this.nBdeApiError := DbiGetCfgMode2Prop(hSession,sesCFGMODE2,nCfgMode2,null,null)

      if (this.nBdeApiError == DBIERR_NONE)                                   // success?

         if ([runtime]$lower(version(0)) = 0)

            ?
            ?[Current Session CFGMODE2 setting:],iif(.not.empty(cProgram),cProgram,[]) at DISPLAY_BDE_COLUMN2
            ?
            ?[nSesCfgMode2 property:],ltrim(str(nCfgMode2)) at DISPLAY_BDE_COLUMN2
            ?[nSesCfgMode2 None:],iif(nCfgMode2 = 0,[Set],[NotSet]) at DISPLAY_BDE_COLUMN2
            ?[nSesCfgMode2 Virtual:],iif(class::IsBitSet(cfgmVirtual,nCfgMode2),[Set],[NotSet]) at DISPLAY_BDE_COLUMN2
            ?[nSesCfgMode2 Persistent:],iif(class::IsBitSet(cfgmPersistent,nCfgMode2),[Set],[NotSet]) at DISPLAY_BDE_COLUMN2
            ?[nSesCfgMode2 Session:],iif(class::IsBitSet(cfgmSession,nCfgMode2),[Set],[NotSet]) at DISPLAY_BDE_COLUMN2
            ?[nSesCfgMode2 All:],iif(class::IsBitSet(cfgmVirtual+cfgmPersistent+cfgmSession,nCfgMode2),[Set],[NotSet]) at DISPLAY_BDE_COLUMN2
            ?

         endif

      else
         class::CheckDbi((this.nBdeApiError),[DbiGetCfgMode2Prop])
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function setCurrentSession(nSessionHandle)

   /*

      Method to set the current BDE session.

   */

      local nSessionHandle

      class::resetBdeApiError()

      this.nBdeApiError := DbiSetCurrSession(nSessionHandle)                  // set current session handle

      if .not.(this.nBdeApiError == DBIERR_NONE)                              // error?
         class::CheckDbi((this.nBdeApiError),[DbiSetCurrSession])             // save it
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function getDatabaseFolders(nDatabaseHandle,cDefault,cWorking)

   /*

      Method to get a current session BDE database default and working directories.

      Notes:

         The database default directory is the application startup directory.
         The database working directory is the application current directory.

   */

      local o,bResult

      bResult = false

      private cDefaultFolder,cWorkingFolder

      cDefaultFolder = ""
      cWorkingFolder = ""

      o = new BdeSession((this))
      cDefaultFolder = o.GetDatabaseDirectory(nDatabaseHandle,true)
      if .not.empty(cDefaultFolder)
         cDefault = cDefaultFolder
         cWorkingFolder = o.GetDatabaseDirectory(nDatabaseHandle,false)
         if .not.empty(cWorkingFolder)
            cWorking = cWorkingFolder
            bResult = true
         endif
      endif
      o.Release()

      return (bResult)

// Method ======================================================================

   function getSessionDatabases()

      local o,nDatabases

      o = new BdeSession((this))
      nDatabases = o.GetDatabases()
      o.Release()

      return (nDatabases)

// Method ======================================================================

   function getSessionHandle()

      local o,nHandle

      o = new BdeSession((this))
      nHandle = o.GetSessionHandle()
      o.Release()

      return (nHandle)

// Method ======================================================================

   function getSessionId()

      local o,nId

      o = new BdeSession((this))
      nId = o.GetId()
      o.Release()

      return (nId)

// Method ======================================================================

   function getSessionNetworkDirectory()

      local o,cNetDir

      o = new BdeSession((this))
      cNetDir = o.GetNetDir()
      o.Release()

      return (cNetDir)

// Method ======================================================================

   function setDatabaseWorkingDirectory(nDatabaseHandle,cFullPath)

   /*

      Method to set a BDE database working directory.

      Notes:

         The database default directory is the application startup directory.
         The database working directory is the application current directory.

   */

      local o,bResult

      bResult = false

      private cWorkingFolder

      cWorkingFolder = trim(cFullPath)

      o = new BdeSession((this))

      if o.SetDatabaseDirectory(nDatabaseHandle,cWorkingFolder)
         bResult = true
      endif

      o.Release()

      return (bResult)

// Method ======================================================================

   function saveSessionDirectoryContext(lResetPrivateDirectory,lDeletePrivateDirectory)

      local lReturn,lTemporaryObjectFolderDelete

      lReturn = true

      class::resetStatus()

      // verify session directory context is NOT saved

      if (.not.empty(this.cSessionNetworkDirectorySaved)).and.(.not.empty(this.cSessionPrivateDirectorySaved))
         class::setStatus([error],[The BDE session directory context has already been saved])
         lReturn := false
      endif

      // save session directory context

      if lReturn

         this.cSessionNetworkDirectorySaved := class::getSessionNetworkDirectory()

         if empty(this.cSessionNetworkDirectorySaved)
            this.cSessionNetworkDirectory := [none]
         endif

         this.cSessionPrivateDirectorySaved := class::getSessionPrivateDirectory()

         if empty(this.cSessionPrivateDirectorySaved)
            this.cSessionPrivateDirectory := [none]
         endif

      endif

      // finish

      if .not.lReturn
         class::showError(program())
      endif

      return (lReturn)

// Method ======================================================================

   function restoreSessionDirectoryContext(lDeletePrivateDirectory)

      local lReturn,lTemporaryObjectFolderDelete

      lReturn = true

      class::resetStatus()

      // process restore

      do case
         case  empty(this.cSessionNetworkDirectorySaved).and.empty(this.cSessionPrivateDirectorySaved)   // not saved
               class::setStatus([error],[The BDE session directory context has NOT been saved])
               lReturn := false
         otherwise

               // reset current private directory
/*
               do case
                  case  empty(class::getSessionPrivateDirectory())            // no current private directory
                  otherwise

                        if argcount() >= 1
                           lTemporaryObjectFolderDelete = lDeletePrivateDirectory
                           class::setSessionPrivateDirectoryDelete(lDeletePrivateDirectory)
                        endif

                        lReturn := class::setSessionPrivateDirectory([])

                        if argcount() >= 1
                           class::setSessionPrivateDirectoryDelete(lTemporaryObjectFolderDelete)
                        endif

               endcase
*/
               // restore network directory

               if lReturn
                  if .not.(lower(this.cSessionNetworkDirectorySaved) == lower(class::getSessionNetworkDirectory()))
                     lReturn := class::setSessionNetworkDirectory((this.cSessionNetworkDirectorySaved))
                  endif
               endif

               // restore private directory

               if lReturn
                  if .not.(lower(this.cSessionPrivateDirectorySaved) == [none])
*                    if class::setSessionPrivateDirectory((this.cSessionPrivateDirectorySaved))
                     if class::setSessionPrivateDirectory()
                        this.lTemporaryObjectsFolderSet := false              // reset
                     else
                        lReturn := false
                     endif
                  endif
               endif

      endcase

      // finish

      if .not.lReturn
         class::showError(program())
      endif

      return (lReturn)

// Method ======================================================================

   function setSessionNetworkDirectory(cFullPath)

   /*

      Method to set the current BDE session network directory (NETDIR) path
      AND creates the NETDIR folder if it does not exist.

   */

      local o,bResult,lAbort,cPath,;
            cNetworkCurrentDirPath,cNetNewDirPath,cPrivateCurrentDirPath

      // initialise

      if class::IsAdministratorAccount()
         return false                                                         // return error
      endif

      cPath = iif(.not.empty(cFullPath),trim(cFullPath),[])

      bResult = false
      lAbort = false

      cNetworkCurrentDirPath = []
      cNetworkNewDirPath = []
      cPrivateCurrentDirPath = []

      class::resetStatus()

      // parse argument

      do case
/*        case class::getSessionId() <> 1                                     // not default session
               this.cErrorMessage = "Multiple BDE sessions are NOT supported"
               lAbort = true
*/
         case  argcount() = 0
               this.cErrorMessage = "A NetworkDirectory absolute path MUST be provided"
               lAbort = true
         otherwise

               do case
                  case  empty(cPath)
                        this.cErrorMessage = "A NetworkDirectory absolute path MUST be provided"
                        lAbort = true
/*
                  case  lower(trim(class::GetSessionNetworkDirectory())) == lower(trim(cPath))  // match?
                        class::messageInformation([The same BdeNetworkDirectory is ALREADY set for the session.],program())
                        bResult := true                                       // already done
*/
                  otherwise

                        cNetworkCurrentDirPath := class::getSessionNetworkDirectory()
                        cPrivateCurrentDirPath := class::getSessionPrivateDirectory()

                        do case
                           case  empty(cPrivateCurrentDirPath)                // no current PrivateDirectory
                                 cNetworkNewDirPath := trim(cPath)
                           otherwise
/*
                                 this.cErrorMessage = "A PrivateDirectory has already been set"
                                 lAbort = true
*/
                                 cNetworkNewDirPath := trim(cPath)
                        endcase

               endcase
      endcase

      // process call

      if .not.lAbort                                                          // ok?
         if .not.class::ActiveDataObjects([messages])                         // no active data objects?

            o = new BdeSession((this))

            if o.SetNetDir(trim(cNetworkNewDirPath))                          // set new NETDIR
               bResult = true
            else
               this.cErrorMessage := o.cErrMsg
            endif

            o.Release()

            if bResult
               if .not.class::DoesAliasPathExist(trim(cNetworkNewDirPath))
                  class::CreateAliasPath(trim(cNetworkNewDirPath))            // create new NetworkDirectory path
               endif
            endif

         else
            lAbort = true
         endif
      endif

      // manage errors

      if lAbort                                                               // need to abort?
         if type("_app.eAbortException") == "O".and._app.eAbortException.baseClassName == "EXCEPTION"
            _app.eAbortException.message := this.cErrorMessage
            _app.eAbortException.cFunction := "ccs_Bde::SetSessionNetDirPath()"
            throw _app.eAbortException                                        // abort
         else
            if this.lBugMessages                                              // issue bug message?
               msgbox("BUG - "+this.cErrorMessage+".",program(),16)
               if ("runtime"$lower(version(0))) = 0
                  if this.lDebugger
                     set step on
                  endif
                  ?program()+" - BUG - "+this.cErrorMessage
                  ?
               endif
            endif
         endif
      endif

      // finish

      this.cStatus := iif(.not.empty(this.cErrorMessage),[error],[ok])        // return class status

      return (bResult)

// Method ======================================================================

   function getSessionPrivateDirectory()

      local o,cPrivDir

      o = new BdeSession((this))
      cPrivDir = o.GetPrivDir()
      o.Release()

      return (cPrivDir)

// Method ======================================================================

   function getSessionPrivateDirectoryDelete()
      return (this.lTemporaryObjectsFolderDelete)

// Method ======================================================================

   function setSessionPrivateDirectoryDelete(lState)
      this.lTemporaryObjectsFolderDelete := lState
      return

// Method ======================================================================

   function setSessionPrivateDirectory(cFullPath)

   /*

      Method to set/reset the BDE session private directory.

      Input:

         cFullPath   =  private directory path to set, NULL to reset to default

      Output:

         lReturn     =  true/false for success/error

      Notes:

         cFullPath is created if needed.

         The database default directory is the application startup directory.
         The database working directory is the application current directory.

         This method is potentially VERY dangerous and MUST be called BEFORE any
         database or query objects are activated. It will check for any active
         data objects, other than a session default database, and return false if
         any are found.

         dBase application initialisation sets the BDE private and working
         directories to the application startup folder. dBase also sets the
         default session default database (_app.databases[1].handle) working
         directory to the application startup folder. This enables the dBase
         navigator and applications to access tables, without using a BDE alias,
         in the current directory.

         DbiSetPrivateDir() sets the working directory for the default
         database to the same folder as the private directory. This change
         will PREVENT the dBase navigator and applications from accessing tables
         without using a BDE alias.

         A DbiSetPrivateDir(null) reset sets the default database working directory
         to the default startup directory. For PLUS.exe this is the PLUS.exe
         startup folder "C:\Program Files\dBase\BIN" and for PLUSrun.exe it is
         where ever PLUSrun.exe started.

         This can produce unpredictable results in the IDE and runtime.

         To alleviate this problem the default session default database working
         directory is set to set("directory") when DbiSetPrivateDir(null) is
         used to reset the BDE private directory. This strategy returns the BDE
         working directory to the same as the dBase application initialisation.

         There are some subtle implications to setting the BDE private directory.

         When DbiSetPrivateDir() sets the private directory it sets the default
         database working directory to the same as the private directory.

         ANY FURTHER ACCESS TO TABLES IN THE APPLICATION STARTUP FOLDER IS NOT
         POSSIBLE UNLESS A BDE ALIAS OR AN ABSOLUTE PATH IS USED TO ACCESS THEM.

   */

      local o,lReturn,cPath,;
            pCurrentDirectory,lPrivateDirectoryReset,lDeletePrivateDirectory

      // initialise

      if class::IsAdministratorAccount()
         return false                                                         // return error
      endif

      cPath = iif(.not.empty(cFullPath),trim(cFullPath),[])
      pCurrentDirectory = []
      lPrivateDirectoryReset = false
      lDeletePrivateDirectory = false

      lReturn = true

      class::resetStatus()

      // check validity

      do case
/*
        case   class::getSessionId() <> 1                                     // not default session
               this.cErrorMessage = "Multiple BDE sessions are NOT supported"
               lReturn := false
*/
         otherwise
               if lower(trim(class::GetSessionPrivateDirectory())) == lower(trim(cPath))  // match?
                  class::messageAlert([The same BdePrivateDirectory is ALREADY set for the session.],program())
               endif
      endcase

      // process command

      if lReturn
         if .not.class::ActiveDataObjects([messages])                         // no active data objects?

            if .not.empty(class::GetSessionNetworkDirectory())                // NetDir set?

               // create private directory, if needed, before setting the private directory

               if .not.empty(cPath)
                  if .not.class::CreateAliasPath(cPath)
                     class::setStatus([error],[Cannot create private directory folder.]+chr(13)+chr(13)+"[ "+cPath+" ]")
                     lReturn := false
                  endif
               endif

               // set private directory

               if lReturn

                  o = new BdeSession((this))

                  pCurrentDirectory := o.GetPrivDir()                         // get session current private directory

                  if o.SetPrivateDirectory(cPath)                             // set session private directory

                     // process private directory set

                     do case

                        // session private directory HAS been set

                        case  this.lTemporaryObjectsFolderSet

                           do case

                              // private directory reset

                              case  empty(cPath)

                                    lDeletePrivateDirectory := true           // delete saved private directory folder
                                    lPrivateDirectoryReset := true            // set private directory reset
                                    this.cTemporaryObjectsFolder := []        // reset saved private directory
                                    this.lTemporaryObjectsFolderSet := false  // set private directory not set

                              // private directory change

                              otherwise

                                    // >>> put a check in for default private directory in v2.70

                                    lDeletePrivateDirectory := true           // delete saved private directory folder
                                    this.cTemporaryObjectsFolder := cPath     // save new private directory
                                    this.lTemporaryObjectsFolderSet := true   // set private directory set

                           endcase

                        // session private directory has NOT been set

                        otherwise

                           do case

                              // private directory set to session configuration default

                              case  empty(cPath)

                                    class::setStatus([error],[No private directory folder provided])
                                    class::messageAlert(this.cErrorMessage+[.],program())
                                    lPrivateDirectoryReset := true            // set private directory reset

                              // new private directory set

                              otherwise

                                    // >>> put a check in for default private directory in v2.70

                                    this.cTemporaryObjectsFolder := cPath     // save new private directory
                                    this.lTemporaryObjectsFolderSet := true   // set private directory set

                           endcase

                     endcase

                     // reset default database working directory and delete saved private directory folder

                     if lReturn

                        // reset default database working directory

                        if lPrivateDirectoryReset                             // reset default database working directory?

                           this.nBdeApiError := DbiSetDirectory(_app.databases[1].handle,set("directory"))

                           if .not.(this.nBdeApiError = DBIERR_NONE)
                              class::CheckDbi((this.nBdeApiError),"DbiSetDirectory")
                              lReturn := false
                           endif

                        endif

                        // delete saved private directory folder

                        if lReturn
                           if lDeletePrivateDirectory                         // delete saved private directory folder?
                              lReturn := class::temporaryObjectsFolderDelete(pCurrentDirectory,[messages])
                           endif
                        endif

                     endif

                  else
                     class::setStatus([error],o.cErrMsg)
                     lReturn := false
                  endif

                  o.Release()

               endif

            else
               class::setStatus([error],[A NetworkDirectory MUST be set before SetSessionPrivateDirectory() can be used])
               lReturn := false
            endif

         else
            lReturn := false
         endif
      endif

      // finish

      if .not.lReturn
         class::showError(program())
      endif

      return (lReturn)

// Method ======================================================================

   function checkTemporaryMemoFiles()

   /*

      Method to check for any _QSQL*.DBT files in a PrivateDirectory folder.

      Input:

         this.cTemporaryObjectsFolder  =  PrivateDirectory full folder path

      Output:

         lReturn                       =  true/false if any found/not found
         this.lTemporaryMemoFiles      =  lReturn

   */

      #define ARRAY_DIR_NAME 1                                                // constants for columns returned by aFiles.dir()
      #define ARRAY_DIR_SIZE 2
      #define ARRAY_DIR_DATE 3
      #define ARRAY_DIR_TIME 4
      #define ARRAY_DIR_ATTR 5

      local lReturn,aFiles,nFiles,i

      lReturn = false
      nFiles = 0

      this.lTemporaryMemoFiles := false                                       // reset class property

      if class::IsPrivateDirectorySet()                                       // BdePrivateDirectory set?
         aFiles = new array()                                                 // create folder files array
         nFiles = aFiles.dir(this.cTemporaryObjectsFolder+[\_QSQL*.DBT],[HS]) // get temporary memo files
         lReturn := (nFiles > 0)                                              // set return
      endif

      this.lTemporaryMemoFiles := lReturn                                     // update class property

      return (lReturn)

// Method ======================================================================

   function temporaryObjectsFolderDelete(cFolder,cMessage)

   /*

      Method to delete the BDE temporary objects folder.

   */

      local lReturn

      lReturn = true

      private  aFiles,nFiles,nFile,cDirectory,nResult

      nResult = 0

      try

         if .not.empty(cFolder)                                               // get a folder?

            if class::getSessionPrivateDirectoryDelete()                      // delete private directory folder set?

               cDirectory = trim(cFolder)

               aFiles = new array()
               nFiles = aFiles.dir(cDirectory+[\*.*],[HS])

               for nFile = 1 to nFiles
                new file().delete(cDirectory+[\]+aFiles[nFile,1])
               endfor

               if .not.RemoveDirectoryA(cDirectory)
                  if .not.empty(cMessage)
                     msgbox([Cannot remove BdePrivateDirectory folder]+chr(13)+chr(13)+" [ "+cDirectory+" ].",program(),48)
                  endif
                  class::setStatus([ok],[Cannot remove BdePrivateDirectory folder])
                  lReturn := false
               endif

               release object aFiles
               aFiles = ""

            endif

         endif

      catch (exception e)

         class::setStatus([error],[Cannot delete files in]+" [ "+cDirectory+" ]."+chr(13)+chr(13)+"[ "+e.message+" ]")
         lReturn := false

      endtry

      release  aFiles,nFiles,nFile,cDirectory,nResult

      return (lReturn)

// Method ======================================================================

   function fixDbiDeleteAliasBug()

   /*

      Method to perform the dBPLUS v2.7 workaround for the DbiDeleteAlias() BDE
      api for persistent aliases.

   */

      local lReturn

      lReturn = true                                                          // assume success

      if this.lFixDbIDeleteAliasBugEnabled                                    // fix v2.7 DbiDeleteAlias() bug?
         if class::isDbaseUACSupportEnabled()                                 // UAC enabled?
            class::showSystemConfigurationInformation(program())              // show current BDE system configuration setup
            class::showClientInformation(program())                           // show current BDE client setup
            if class::restartBDE([v2.7])                                      // restart client BDE?
               if class::setDatabaseWorkingDirectory(_app.databases[1].handle,;
                                                     set([directory]))        // reset default session default database working directory
                  if ([runtime]$lower(version(0))) = 0                        // IDE?
                     ?
                     ?[>>> ]+program(),[BDE restarted with v2.7 defaults] at DISPLAY_BDE_COLUMN2
                     ?
                  endif
               else
                  lReturn := false
               endif
            else
               lReturn := false
            endif
            class::showSystemConfigurationInformation(program())              // show restarted BDE system configuration setup
            class::showClientInformation(program())                           // show restarted BDE setup
         endif
      endif

      if .not.lReturn
         class::messageError((this.cMessageError),program())
      endif

      return (lReturn)

// Method ======================================================================

   function restartBDE(cEnvironment,cBdeConfigFileFullPath)

   /*

   Vic McClung

      Rewritten   -  March 2006  by Ronnie MacGregor to return boolean success.
      Enhanced    -  23-Mar-2011 by Andrew Shimmin to return BDE API error.
      Enhanced    -  25-Mar-2011 by Andrew Shimmin to initialise for a v2.70 UAC setup.

   Note:

      The pEnv.szIniFile property used is cBdeConfigFileFullPath if provided.

      If cBdeConfigFileFullPath is NOT provided, the ccs_Bde::cSystemIniFile
      property, set by ccs_Bde::getSystemConfigurationInformation() called by
      ccs_Bde::initialiseObject() at object instantiation, is used.

   */

      local lReturn,pEnv

      lReturn = true                                                          // assume success
      pEnv = []

      if .not.class::ActiveDataObjects([messages])                            // no active data objects?

         class::resetBdeApiError()

         this.nBdeApiError := DbiExit()

         if (this.nBdeApiError == DBIERR_NONE)                                // success?

            do case
               case  empty(cEnvironment).or.(lower(trim(cEnvironment)) == [default])
                     pEnv := null
               case  lower(trim(cEnvironment)) == [v2.7]                      // v2.7 UAC support

                     pEnv := new BdeEnvironmentStructure()

                     pEnv.szWorkDir(_app.currentUserPath)

                     do case
                        case  empty(cBdeConfigFileFullPath)
                              pEnv.szIniFile((this.cSystemIniFile))
                        otherwise
                              if file(trim(cBdeConfigFileFullPath))
                                 pEnv.szIniFile(trim(cBdeConfigFileFullPath))
                              else
                                 pEnv.szIniFile((this.cSystemIniFile))
                              endif
                     endcase

                     pEnv.bForceLocalInit((this.lClientEnvForceLocalInit))
                     pEnv.szLang(this.cClientLanguage)
                     pEnv.szClientName([BdeTesting_v2.7])

               otherwise
                     pEnv := null
            endcase

            this.nBdeApiError := DbiInitFn(DBIINTFVER,pEnv)

            if .not.(this.nBdeApiError == DBIERR_NONE)                        // error?
               class::CheckDbi((this.nBdeApiError),[DbiInitFn])               // save error
            endif

            if .not.empty(pEnv)
               pEnv.release()
            endif

         else
            class::CheckDbi((this.nBdeApiError),[DbiExit])                    // save error
         endif

         lReturn := (this.nBdeApiError == DBIERR_NONE)

      else
         lReturn := false                                                     // active data objects
      endif

      return (lReturn)

// Method ======================================================================

   function createAlias(cAliasName,cFullPath,cTableFormat)

      class::resetBdeApiError()

      if argCount() < 3
         cTableFormat = [DBASE]
      endif

      this.nBdeApiError := DbiAddAlias(null,;
                                       oem(cAliasName),;
                                       oem(cTableFormat),;
                                       oem([PATH:]+cFullPath),;
                                       TRUE)                                  // add alias

      if .not.(this.nBdeApiError == DBIERR_NONE)                              // error?
         class::CheckDbi((this.nBdeApiError),[DbiAddAlias])                   // save it
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function createDbaseAlias(cAliasName,cFullPath)

      class::resetBdeApiError()

      this.nBdeApiError := DbiAddAlias(null,;
                                       oem(cAliasName),;
                                       oem([DBASE]),;
                                       oem([PATH:]+cFullPath),;
                                       TRUE)                                  // create alias

      if .not.(this.nBdeApiError == DBIERR_NONE)                              // error?
         class::CheckDbi((this.nBdeApiError),[DbiAddAlias])                   // save it
      endif

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function deleteAlias(cAliasName)
   /*

      Method to delete a PERSISTENT alias ONLY.

      Input:

         cAliasName  =  BDE persistent alias to delete

      Output:

         true/false for success/failure

         this.nBdeApiError =  BDE error code returned to the api
         this.cBdeApiError =  BDE error string for the BDE error code

      Notes:

         dBPLUS v2.70 ( > b2119 ) provides support for the Microsoft security model.
         dBPLUS v2.70 also added session object methods for creating/deleting
         NON-PERSISTENT aliases.

         The session methods for non-persistent aliases broke the DbiDeleteAlias()
         BDE api for PERSISTENT aliases and, although returning success when deleting
         a persistent alias, the alias is NOT deleted.

         The fix is to set the session sesCFGMODE2 property correctly before
         calling the DbiDeleteAlias() api and then reset it following the call to
         DbiDeleteAlias().

         This method is to be USED to delete ONLY persistent BDE aliases.

   */

      local hSession,nSesCfgMode2,lReset

      // initialise

      hSession = 0
      nSesCfgMode2 = 0
      lReset = true

      class::resetBdeApiError()

      hSession := class::getSessionHandle()

      // delete alias

      do case
         case  class::IsDbaseUACSupported()                                   // >= v2.7

               if class::GetCurrentSessionHandle(hSession)

                  // set current sesCFGMODE2 property for v2.7

                  if class::getSesCfgMode2(nSesCfgMode2)

                     this.nBdeApiError := DbiSetProp(hSession,sesCFGMODE2,cfgmPersistent)

                     if (this.nBdeApiError == DBIERR_NONE)

                        // delete persistent alias

                        this.nBdeApiError := DbiDeleteAlias(null,oem(cAliasName))
                        if .not.(this.nBdeApiError == DBIERR_NONE)
                           class::CheckDbi((this.nBdeApiError),[DbiDeleteAlias])
                        endif

                        // reset sesCFGMODE2 property for v2.7 - even on a DbiDeleteAlias() error

                        this.nBdeApiError := DbiSetProp(hSession,sesCFGMODE2,nSesCfgMode2)
                        if .not.(this.nBdeApiError == DBIERR_NONE)
                           class::CheckDbi((this.nBdeApiError),[DbiSetProp])
                        endif

                     else
                        class::CheckDbi((this.nBdeApiError),[DbiSetProp])
                     endif

                  endif

               endif

         otherwise                                                            // < v2.7

               // delete persistent alias

               this.nBdeApiError := DbiDeleteAlias(null,oem(cAliasName))

               if .not.(this.nBdeApiError == DBIERR_NONE)
                  class::CheckDbi((this.nBdeApiError),[DbiDeleteAlias])
               endif

      endcase

      // finish

      return (this.nBdeApiError == DBIERR_NONE)

// Method ======================================================================

   function checkDbi(iResult,cBdeApi)

      local    iResult

      private  cApi

      this.nBdeApiError := iResult                                            // reset
      this.cBdeApiError := []

      if .not.(iResult = 0)

         do case
            case  argcount() < 2
                  cApi = ""
            otherwise
                  cApi = trim(argvector(2))
         endcase

         this.cBdeApiError = space(DBIMAXMSGLEN+1)
         DbiGetErrorString(iResult,this.cBdeApiError)

         msgbox([BDE ]+iif(empty(cApi),[ ],cApi+[() ])+[API error.]+chr(13)+chr(13)+;
                "[ "+[Error: ]+ltrim(str(iResult))+[ - ]+this.cBdeApiError+" ]",program(),16)

         if ([runtime]$lower(version(0))) = 0
            ?program()+[ - BUG - BDE ]+iif(empty(cApi),[ ],cApi+[() ])+[API error]
            ?program()+[ - BUG - ]+"[ "+[Error: ]+ltrim(str(iResult))+[ - ]+this.cBdeApiError+" ]"
            ?
            if this.lDebugger
               set step on
            endif
         endif

      endif

      release cApi

      return (iResult = 0)

/*
   =============================================================================

   class messaging methods

   =============================================================================
*/

// Method ======================================================================

   function messageAlert(cMessage,cTitle)
      return class::messageShow(cMessage,cTitle,M_MB_ICONEXCLAMATION)

// Method ======================================================================

   function messageError(cMessage,cTitle)
      return class::messageShow(cMessage,cTitle,M_MB_ICONHAND)

// Method ======================================================================

   function messageInformation(cMessage,cTitle)
      return class::messageShow(cMessage,cTitle,M_MB_ICONINFORMATION)

// Method ======================================================================

   function messageYesNo(cMessage,cTitle)
      return class::messageShow(cMessage,cTitle,M_MB_ICONQUESTION+M_MB_YESNO)

// Method ======================================================================

   function messageShow(cMessage,cTitle,nIcon)
      return   MessageBox(this.hMessageParentWindow,cMessage,cTitle,M_MB_OK+nIcon+M_MB_TASKMODAL+M_MB_TOPMOST+M_MB_SETFOREGROUND)

// Method ======================================================================

   function messageSetParentWindow(nHandle)

   /*

      Method to set the parent window for the messaging methods.

   */

      if argcount() = 0
         this.hMessageParentWindow := 0
      else
         this.hMessageParentWindow := nHandle
      endif

      return

// Method ======================================================================

   function messageException(e,cTitle,cLogIdentifier,cStatus)

   /*

      Function to display an exception message

      Input:

         e              =  exception object reference
         cTitle         =  title of msgbox window
         cLogIdentifier =  log exception identifier (application, function or method)
         cStatus        =  status at exception

      Output:

         A modal msgbox() is displayed using the function arguments

   */

      local cId

      cId = iif(.not.empty(cStatus),trim(cStatus),[NotProvided])

      class::messageError(e.message+chr(13)+chr(13)+;
                          [File: ]+e.fileName+chr(13)+chr(13)+;
                          [Line: ]+ltrim(str(e.lineNo))+chr(13)+chr(13)+;
                          [State: ]+cId,;
                          cTitle)                                             // display user message

      if argcount() >= 3                                                      // log message?
         if type([_log]) == [FP]                                              // logging available?
            _log()
            _log([>>> ]+cLogIdentifier+[ EXCEPTION])
            _log()
            _log([Message],e.message)
            _log([File],e.fileName)
            _log([Line],ltrim(str(e.lineNo)))
            _log([State],cId)
            _log()
         endif
      endif

      return

endclass

Class BdeConfiguration(lDebugger) of object custom
/*
   =============================================================================

   Class constructor

   =============================================================================
*/

  // custom properties

   this.baseClassName = [OBJECT]
   this.className = [BDECONFIGURATIONBJECT]

   this.error  = 0
   this.cErrMsg = []
   this.lDebugger = lDebugger                                                 // set debugger

   this.hCfg = null                                                           // handle for DbiOpenCfgInfoList()
   this.phCursor = new BdeHandlePointer()                                     // handle pointer structure for DbiOpenCfgInfoList() returned cursor
   this.sCFGDesc = new BdeCFGDescStructure()                                  // CFGDesc structure for DbiGetNextRecord()
   this.sRECProps = new BdeRECPropsStructure()                                // RECProps structure for DbiGetNextRecord()

   this.nDbaseTableLevel = 0                                                  // current configuration file dBase driver table level
   this.cDbaseLanguageDriver = []                                             // current configuration file dBase driver language driver

/*
   =============================================================================

   Class methods

   =============================================================================
*/

// Method ======================================================================

   function CheckDbi(iResult,cBdeApi)

      local    lResult

      private  cApi

      if .not.(iResult = 0)

         do case
            case  argcount() < 2
                  cApi = ""
            otherwise
                  cApi = trim(argvector(2))
         endcase

         this.cErrMsg = space(DBIMAXMSGLEN+1)
         DbiGetErrorString(iResult,this.cErrMsg)

         msgbox("BDE "+iif(empty(cApi)," ",cApi+"() ")+"API error."+chr(13)+chr(13)+;
                "[ Error: "+ltrim(str(iResult))+" - "+this.cErrMsg+" ]",program(),16)

         if ("runtime"$lower(version(0))) = 0
            ?program()+" - BUG - BDE "+iif(empty(cApi)," ",cApi+"() ")+"API error"
            ?program()+" - BUG - [ Error: "+ltrim(str(iResult))+" - "+this.cErrMsg+" ]"
            ?
            if this.lDebugger
               set step on
            endif
         endif

      endif

      return (iResult = 0)

// Method ======================================================================

   function GetErrorCode()
      return (this.error)

// Method ======================================================================

   function GetSetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

   /*

      Method to get a setting from the current BDE configuration file.

      Input:

         cBdeConfigPath    =  BDE configuration path              eg. [\DRIVERS\DBASE\TABLE CREATE]
         cNodeName         =  BDE configuration path node name    eg. [LEVEL]

         cDisplayResults   =  .not.empty() to display results in the command pane

      Output:

         bResult           =  true/false for success/error
         cNodeValue        =  BDE configuration node value        eg. [7]

      Notes:

         All node values returned are strings. Boolean node values are returned
         as [TRUE]/[FALSE].

         To get all BDE configuration paths available use:

            cBdeConfigPath = [\]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

         To get each top level configuration path node names use eg.:

            cBdeConfigPath = [\Databases]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

   */

      local bResult,hCursor,lDisplay

      // initialise

      bResult = true
      hCursor = 0
      cNodeValue = []
      lDisplay = (.not.empty(cDisplayResults)).and.(([runtime]$lower(version(0))) = 0)

      // open BDE configuration file path

      this.error := DbiOpenCfgInfoList(this.hCfg,dbiREADONLY,cfgPersistent,trim(cBdeConfigPath),this.phCursor)

      // get configuration path node value

      if this.error = DBIERR_NONE

         hCursor := this.phCursor.phCur()                                     // get configuration path cursor handle
         this.error := DbiSetToBegin(hCursor)                                 // start at beginning

         if this.error = DBIERR_NONE

            if lDisplay
               ?
               ?[BDE current configuration databases:]
               ?
            endif

            // check all configuration path nodes

            do while this.error = DBIERR_NONE

               this.error := DbiGetNextRecord(hCursor,dbiNOLOCK,this.sCFGDesc,this.sRECProps)

               if this.error = DBIERR_NONE

                  if lDisplay
                     ?[NodeName: ]+this.sCFGDesc.szNodeName(),[Value: ]+this.sCFGDesc.szValue() at DISPLAY_BDE_COLUMN2
                  endif

                  do case
                     case  upper(this.sCFGDesc.szNodeName()) == upper(trim(cNodeName))
                           cNodeValue= this.sCFGDesc.szValue()
                  endcase

               else

                  do case
                     case  this.error = DBIERR_EOF                            // no more
                     otherwise
                           bResult := class::CheckDbi((this.error),[DbiGetNextRecord])
                  endcase

               endif

            enddo

            this.error := iif(this.error = DBIERR_EOF,DBIERR_NONE,this.error) // return success at end of nodes

         else
            bResult := class::CheckDbi((this.error),[DbiSetToBegin])
         endif

         // close configuration path cursor

         if .not.empty(hCursor)
            this.error := DbiCloseCursor(this.phCursor)
            if this.error <> DBIERR_NONE
               bResult := class::CheckDbi((this.error),[DbiCloseCursor])
            endif
         endif

      else
         bResult := class::CheckDbi((this.error),[DbiOpenCfgInfoList])
      endif

      // finish

      return (bResult)

// Method ======================================================================

   function ModifySetting(cBdeConfigPath,cNodeName,cNodeValue,cDisplayResults)

   /*

      Method to modify a setting from the current BDE configuration file.

      Input:

         cBdeConfigPath    =  BDE configuration path              eg. [\SYSTEM\INIT]
         cNodeName         =  BDE configuration path node name    eg. [MAXFILEHANDLES]
         cNodeValue        =  BDE configuration node value        eg. [1024]

         cDisplayResults   =  .not.empty() to display results in the command pane

      Output:

         bResult           =  true/false for success/error

      Notes:

         End nodes are called leaf nodes. ONLY leaf nodes can be modified.

         All node values returned are strings. Boolean node values are returned
         as [TRUE]/[FALSE].

         To get all BDE configuration paths available use:

            cBdeConfigPath = [\]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

         To get each top level configuration path node names use eg.:

            cBdeConfigPath = [\Databases]
            cNodeValue = []
            cDisplayResults = [display]

         from the IDE.

   */

      local bResult,hCursor,lDisplay,lDone

      // initialise

      bResult = true
      hCursor = 0
      lDisplay = (.not.empty(cDisplayResults)).and.(([runtime]$lower(version(0))) = 0)
      lDone = false

      // open BDE configuration file path

      this.error := DbiOpenCfgInfoList(this.hCfg,dbiREADWRITE,cfgPersistent,trim(cBdeConfigPath),this.phCursor)

      // get configuration path node value

      if this.error = DBIERR_NONE

         hCursor := this.phCursor.phCur()                                     // get configuration path cursor handle
         this.error := DbiSetToBegin(hCursor)                                 // start at beginning

         if this.error = DBIERR_NONE

            if lDisplay
               ?
               ?[BDE current configuration databases:]
               ?
            endif

            // check all configuration path nodes

            do while this.error = DBIERR_NONE.and.(.not.lDone)

               this.error := DbiGetNextRecord(hCursor,dbiNOLOCK,this.sCFGDesc,this.sRECProps)

               if this.error = DBIERR_NONE

                  if lDisplay
                     ?[NodeName: ]+this.sCFGDesc.szNodeName(),[Value: ]+this.sCFGDesc.szValue() at DISPLAY_BDE_COLUMN2
                  endif

                  do case
                     case  upper(this.sCFGDesc.szNodeName()) == upper(trim(cNodeName))

                           this.sCFGDesc.szValue(trim(cNodeValue))
                           this.error := DbiModifyRecord(hCursor,this.sCFGDesc,true)   // modify configuration record
                           bResult := class::CheckDbi((this.error),[DbiModifyRecord])

                           lDone := true

                  endcase

               else

                  do case
                     case  this.error = DBIERR_EOF                            // no more
                     otherwise
                           bResult := class::CheckDbi((this.error),[DbiGetNextRecord])
                  endcase

               endif

            enddo

            this.error := iif(this.error = DBIERR_EOF,DBIERR_NONE,this.error) // return success at end of nodes

         else
            bResult := class::CheckDbi((this.error),[DbiSetToBegin])
         endif

         // close configuration path cursor

         if .not.empty(hCursor)
            this.error := DbiCloseCursor(this.phCursor)
            if this.error <> DBIERR_NONE
               bResult := class::CheckDbi((this.error),[DbiCloseCursor])
            endif
         endif

      else
         bResult := class::CheckDbi((this.error),[DbiOpenCfgInfoList])
      endif

      // finish

      return (bResult)

// Method ======================================================================

   function Release()

      if type([this.sCFGDesc]) # [U]
         this.sCFGDesc.release()
         this.sCFGDesc := null
      endif
      if type([this.sRECProps]) # [U]
         this.sRECProps.release()
         this.sRECProps := null
      endif
      if type([this.phCursor]) # [U]
         this.phCursor.release()
         this.phCursor := null
      endif

      this.hCfg := null

      try
         release object this
      catch (exception e)
      endtry

      return

endclass

Class BdeSession(oBde) of object custom
/*
   =============================================================================

   Class constructor

   =============================================================================
*/

  // custom properties

   this.baseClassName = [OBJECT]
   this.className = [BDESESSIONOBJECT]

   this.oBde = oBde

   this.error  = 0
   this.cErrMsg = []

   this.hSession = 0
   this.sSession = new BdeSessionStructure()
   this.nId = 0
   this.cName = []
   this.nDatabases = 0
   this.nCursors = 0
   this.nLockWait = 0
   this.cNetDir = []
   this.cPrivDir = []

   this.cDatabaseDefaultDirectory = []
   this.cDatabaseWorkingDirectory = []

   class::GetSessionInformation()                                             // get BdeSession information

/*
   =============================================================================

   Class methods

   =============================================================================
*/

// Method ======================================================================

   function CheckDbi(iResult,cBdeApi)

      local    lResult

      private  cApi

      if .not.(iResult = 0)

         do case
            case  argcount() < 2
                  cApi = ""
            otherwise
                  cApi = trim(argvector(2))
         endcase

         this.cErrMsg = space(DBIMAXMSGLEN+1)
         DbiGetErrorString(iResult,this.cErrMsg)

         msgbox("BDE "+iif(empty(cApi)," ",cApi+"() ")+"API error."+chr(13)+chr(13)+;
                "[ Error: "+ltrim(str(iResult))+" - "+this.cErrMsg+" ]",program(),16)

         if ("runtime"$lower(version(0))) = 0
            ?program()+" - BUG - BDE "+iif(empty(cApi)," ",cApi+"() ")+"API error"
            ?program()+" - BUG - [ Error: "+ltrim(str(iResult))+" - "+this.cErrMsg+" ]"
            ?
            if this.oBde.lDebugger
               set step on
            endif
         endif

      endif

      return (iResult = 0)

// Method ======================================================================

   function GetDatabaseDirectory(nDatabaseHandle,lType)

      local cDirectory,sDirectory

      cDirectory = ""

      if nDatabaseHandle <> 0

         sDirectory = new BdeDirectoryStructure()

         this.error := DbiGetDirectory(nDatabaseHandle,lType,sDirectory)

         if (this.error = DBIERR_NONE)
            do case
               case  lType
                     this.cDatabaseDefaultDirectory := trim(sDirectory.szDirectory())
                     cDirectory = this.cDatabaseDefaultDirectory
               otherwise
                     this.cDatabaseWorkingDirectory := trim(sDirectory.szDirectory())
                     cDirectory = this.cDatabaseWorkingDirectory
            endcase
         else
            class::CheckDbi((this.error),"DbiGetDirectory")
         endif

         sDirectory.release()

      else
         msgbox("BUG - Database handle not provided.",program(),16)
         if ("runtime"$lower(version(0))) = 0
            ?program()+" - BUG - Database handle not provided"
            ?
            if this.oBde.lDebugger
               set step on
            endif
         endif
      endif

      return (cDirectory)

// Method ======================================================================

   function SetDatabaseDirectory(nDatabaseHandle,cFullPath)
   /*

      Method to set a database object's working directory.

      Notes:

         The database default directory is the application startup directory.
         The database working directory is the application current directory.

   */

      local cDirectory,sDirectory

      cDirectory = trim(cFullPath)

      if nDatabaseHandle <> 0

         this.error := DbiSetDirectory(nDatabaseHandle,cDirectory)

         if .not.(this.error == DBIERR_NONE)
            class::CheckDbi((this.error),"DbiSetDirectory")
         endif

      else
         msgbox("BUG - Database handle not provided.",program(),16)
         if ("runtime"$lower(version(0))) = 0
            ?program()+" - BUG - Database handle not provided"
            ?
            if this.oBde.lDebugger
               set step on
            endif
         endif
      endif

      return (this.error == DBIERR_NONE)

// Method ======================================================================

   function GetSessionInformation()

      local bResult

      bResult = false

      // Get the BDE session

      this.error := DbiGetSesInfo(this.sSession)

      if this.error = DBIERR_NONE

         // Get session properties

         this.nId := this.sSession.iSession()
         this.cName := this.sSession.szName()
         this.nDatabases := this.sSession.iDatabases()
         this.nCursors := this.sSession.iCursors()
         this.nLockWait := this.sSession.iLockWait()
         this.cNetDir := this.sSession.szNetDir()
         this.cPrivDir := this.sSession.szPrivDir()

         class::GetSessionHandle()

         bResult = true

      else

         class::CheckDbi((this.error),"DbiGetSesInfo")

      endif

      return bResult

// Method ======================================================================

   function GetSessionHandle()

      local nHandle

      nHandle = 0

      this.error := DbiGetCurrSession(nHandle)

      if (this.error = DBIERR_NONE)
         this.hSession := nHandle
      else
         class::CheckDbi((this.error),"DbiGetCurrSession")
      endif

      return iif(this.error = DBIERR_NONE,nHandle,0)

// Method ======================================================================

   function GetDatabases()
      return this.nDatabases

// Method ======================================================================

   function GetId()
      return this.nId

// Method ======================================================================

   function GetPrivDir()
      return this.cPrivDir

// Method ======================================================================

   function GetNetDir()
      return this.cNetDir

// Method ======================================================================

   function SetPrivateDirectory(cFullPath)

   /*

      Method to set the private directory.

      Input:

         cFullPath   =  private directory path to set, NULL to reset to default

      Output:

         bResult     =  true/false for success/error

   */

      local bResult,pDirectory

      bResult = true
      pDirectory = iif(empty(cFullPath),NULL,trim(cFullPath))                 // initialise private directory

      this.error := DbiSetPrivateDir(pDirectory)                              // set private directory

      if .not.(this.error = DBIERR_NONE)                                      // error?
         class::CheckDbi((this.error),[DbiSetPrivateDir])
         bResult := false
      endif

      return (bResult)

// Method ======================================================================

   function SetNetDir(cFullPath)

   /*

      Method to set the current BDE session NETDIR path.

   */

      local bResult,nHandle,cNetDirPath

      bResult = false
      nHandle = 0

      nHandle = class::GetSessionHandle()

      if nHandle <> 0

         do case
            case  empty(cFullPath)
                  cNetDirPath = null
            otherwise
                  cNetDirPath = trim(cFullPath)
         endcase

         this.error := DbiSetNetDirProp(this.hSession,sesNETFILE,cNetDirPath)

         if this.error = DBIERR_NONE
            this.hSession := nHandle
            bResult = true
         else
            class::CheckDbi((this.error),[DbiSetNetDirProp])
         endif
      endif

      return (bResult)

// Method ======================================================================

   function Release()

      this.oBde := null
      this.sSession.release()
      release object this

      return

endclass

class BdeSessionStructure of StructBase(563)

/*

   Structure for the BdeSession class.

*/

// Method ======================================================================

   function iSession(c)
      return iif(pcount() = 0,this.GetWord(0),this.SetWord(0,c))

// Method ======================================================================

   function szName(c)
      return iif(pcount() = 0,this.AnsiString(2,DBIMAXNAMELEN+1),;
                              this.AnsiString(2,DBIMAXNAMELEN+1,c))

// Method ======================================================================

   function iDatabases(c)
      return iif(pcount() = 0,this.GetWord(34),this.SetWord(34,c))

// Method ======================================================================

   function iCursors(c)
      return iif(pcount() = 0,this.GetWord(36),this.SetWord(36,c))

// Method ======================================================================

   function iLockWait(c)
      return iif(pcount() = 0,this.GetWord(38),this.SetWord(38,c))

// Method ======================================================================

   function szNetDir(c)
      return iif(pcount() = 0,this.AnsiString(40,DBIMAXPATHLEN+1),;
                              this.AnsiString(40,DBIMAXPATHLEN+1,c))

// Method ======================================================================

   function szPrivDir(c)
      return iif(pcount() = 0,this.AnsiString(301,DBIMAXPATHLEN+1),;
                              this.AnsiString(301,DBIMAXPATHLEN+1,c))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeDirectoryStructure of StructBase(DBIMAXPATHLEN+1)

/*

   Structure for directory string pointer argument.

*/

// Method ======================================================================

   function szDirectory(c)
      return iif(pcount() = 0,this.AnsiString(0,DBIMAXPATHLEN+1),;
                              this.AnsiString(0,DBIMAXPATHLEN+1,c))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeHandlePointer of StructBase(4)

/*

   Structure for a BDE api handle pointer

*/

// Method ======================================================================

   function phCur(n)
      return iif(pcount() = 0,this.member(0),this.member(0,n))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeCFGDescStructure of StructBase(292)

/*

   Structure for dbiOpenCfgInfoList() api.

*/

// Method ======================================================================

   function szNodeName(c)
      return iif(pcount() = 0,this.AnsiString(0,DBIMAXNAMELEN+1),;
                              this.AnsiString(0,DBIMAXNAMELEN+1,c))

// Method ======================================================================

   function szDescription(c)
      return iif(pcount() = 0,this.AnsiString(32,DBIMAXSCFLDLEN),;
                              this.AnsiString(32,DBIMAXSCFLDLEN,c))

// Method ======================================================================

   function iDataType(c)
      return iif(pcount() = 0,this.GetWord(160),this.SetWord(160,c))

// Method ======================================================================

   function szValue(c)
      return iif(pcount() = 0,this.AnsiString(162,DBIMAXSCFLDLEN),;
                              this.AnsiString(162,DBIMAXSCFLDLEN,c))

// Method ======================================================================

   function bHasSubnodes(n)
      return iif(pcount() = 0,this.GetWord(290),this.SetWord(290,n))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeRECPropsStructure of StructBase(14)

/*

   Structure for dbiGetNextRecord() api.

*/

// Method ======================================================================

   function iSeqNum(n)
      return iif(pcount() = 0,this.GetLong(0),this.SetLong(0,n))

// Method ======================================================================

   function iPhyRecNum(n)
      return iif(pcount() = 0,this.GetLong(4),this.SetLong(4,n))

// Method ======================================================================

   function bRecChanged(n)
      return iif(pcount() = 0,this.GetWord(8),this.SetWord(8,n))

// Method ======================================================================

   function bSeqNumChanged(n)
      return iif(pcount() = 0,this.GetWord(10),this.SetWord(10,n))

// Method ======================================================================

   function bDeleteFlag(n)
      return iif(pcount() = 0,this.GetWord(12),this.SetWord(12,n))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeEnvironmentStructure of StructBase(688)

/*

   Structure for the BDE APIs DbiInit() & DbiInitFn().

*/

// Method ======================================================================

   function szWorkDir(c)
      return iif(pcount() = 0,this.AnsiString(0,DBIPATH),;
                              this.AnsiString(0,DBIPATH,c))

// Method ======================================================================

   function szIniFile(c)
      return iif(pcount() = 0,this.AnsiString(261,DBIPATH),;
                              this.AnsiString(261,DBIPATH,c))

// Method ======================================================================

   function bForceLocalInit(l)
      return iif(pcount() = 0,this.GetLogical(622),this.SetLogical(622,l))

// Method ======================================================================

   function szLang(c)
      return iif(pcount() = 0,this.AnsiString(624,DBINAME),;
                              this.AnsiString(624,DBINAME,c))

// Method ======================================================================

   function szClientName(c)
      return iif(pcount() = 0,this.AnsiString(656,DBINAME),;
                              this.AnsiString(656,DBINAME,c))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeClientInfoStructure of StructBase(327)

/*

   Structure for the BDE API DbiGetClientInfo().

*/

// Method ======================================================================

   function szName(c)
      return iif(pcount() = 0,this.AnsiString(0,DBINAME),;
                              this.AnsiString(32,DBINAME,c))

// Method ======================================================================

   function iSessions(n)
      return iif(pcount() = 0,this.GetWord(32),this.SetWord(32,n))

// Method ======================================================================

   function szWorkDir(c)
      return iif(pcount() = 0,this.AnsiString(34,DBIPATH),;
                              this.AnsiString(34,DBIPATH,c))

// Method ======================================================================

   function szLang(c)
      return iif(pcount() = 0,this.AnsiString(295,DBINAME),;
                              this.AnsiString(295,DBINAME,c))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

class BdeSYSConfigStructure of StructBase(575)

/*

   Structure for the BDE API DbiGetSysConfig().

*/

// Method ======================================================================

   function bLocalShare(l)
      return iif(pcount() = 0,this.GetLogical(0),this.SetLogical(0,l))

// Method ======================================================================

   function iNetProtocol(n)
      return iif(pcount() = 0,this.GetWord(2),this.SetWord(2,n))

// Method ======================================================================

   function bNetShare(l)
      return iif(pcount() = 0,this.GetLogical(4),this.SetLogical(4,l))

// Method ======================================================================

   function szNetType(c)
      return iif(pcount() = 0,this.AnsiString(6,DBINAME),;
                              this.AnsiString(6,DBINAME,c))

// Method ======================================================================

   function szUserName(c)
      return iif(pcount() = 0,this.AnsiString(38,DBIUSERNAME),;
                              this.AnsiString(38,DBIUSERNAME,c))

// Method ======================================================================

   function szIniFile(c)
      return iif(pcount() = 0,this.AnsiString(53,DBIPATH),;
                              this.AnsiString(53,DBIPATH,c))

// Method ======================================================================

   function szLangDriver(c)
      return iif(pcount() = 0,this.AnsiString(314,DBINAME),;
                              this.AnsiString(314,DBINAME,c))

// Method ======================================================================

   function release()
      super::release()
      return

endclass

