/*

File: gwGradient.cc
Author: Gary White
Last modified: Feb 10, 2002


   -----------------------------------------------------------
   Note that all code contained in this file is 
   Copyright c2002, Gary White, All Rights Reserved.
   -----------------------------------------------------------
   License:
   You are hereby granted permission to use this code within 
   your own applications, but you are expressly forbidden from 
   selling, or otherwise distributing this source code without 
   prior written consent.
   -----------------------------------------------------------
   Warranty:
   This code freeware and there is no warranty, either express
   or implied. Use this code at your own risk. Your use of this
   code consitutes your agreement with this warranty and its
   license.
   -----------------------------------------------------------


Description:
	Paints a gradient background that fills its container


Written for:
	32-bit versions of dBASE, including VdB 7.x and dB2K


External dependencies: none


Usage:

   In the command window:
   SET PROCEDURE TO GWGRADIENT.CC ADDITIVE

   In the form designer, drag a gwGradient object from the
   Custom page of the component palette to the form, or
   container/notebook that you want to fill with the gradient.
   Note that the gwGradient object must be before (in z-order)
   any other objects that occupy the same container, or the 
   gradient will cover up the other object(s).

   You can use either the form's onOpen event, or an overridden
   open method to set the colors and fadeStyle as desired.

   Example of onOpen handling:

   function onOpen
      with(form.gwGradient1)
         color1 = 0xFF0000 // blue
         color2 = 0x0000FF // red
         fadeStyle = 1 // vertical
      endwith
      return

Also see the form gwGradientDemo.wfm for more examples.

*/

// code block to create an RGBCOLOR value from the
// red, green and blue components
#define RGB(r, g, b) ;
   bitlshift(bitand(b, 0xFF), 16) + ;
   bitlshift(bitand(g, 0xFF), 8) + ;
   bitand(0xFF, r)

// code blocks to extract the individual red, green,
// or blue values from an RGBCOLOR value
#define REDVAL(x) bitand(x, 0x0000FF)
#define GREENVAL(x) bitrshift(bitand(x, 0x00FF00), 8)
#define BLUEVAL(x) bitrshift(bitand(x, 0xFF0000), 16)

// values for GetDeviceCaps
#define PLANES 14
#define BITSPIXEL 12

// Which direction to fade
#define FADE_VERTICAL   1
#define FADE_HORIZONTAL 2

class gwGradient(o) of paintbox(o)
   protect BitsPerPixel, NbrPlanes, ColorBits, init, ;
      onOpen, onPaint, RgnCnt

   // public properties

   // set it small so it won't be in the way in the designer

   this.height = 0
   this.width = 0

   // custom properties

   // which direction to fade
   // FADE_VERTICAL (1) means that it will 
   //    fade from top (color1) to bottom (color2)
   // FADE_HORIZONTAL (2) means that it will 
   //    fade from left (color1) to right (color2)

   this.fadeStyle = FADE_VERTICAL

   // the top, or left color depending on fadeStyle

   this.Color1 = RGB(0,0,255)

   // the bottom, or right color depending on fadeStyle

   this.Color2 = RGB(0,0,0)

   // protected properties used by the class

   this.BitsPerPixel = 0
   this.NbrPlanes = 0
   this.ColorBits = 0

   this.init()

   function init

      // prototype some api calls needed by the class

      if type('BeginPaint') # 'FP'
         extern chandle BeginPaint(chandle, cptr) user32
      endif
      if type('CreateSolidBrush') # 'FP'
         extern chandle CreateSolidBrush(culong) gdi32
      endif
      if type('DeleteObject') # 'FP'
         extern clogical DeleteObject(chandle) gdi32
      endif
      if type('EndPaint') # 'FP'
         extern cvoid EndPaint(chandle, cptr) user32
      endif 
      if type('GetClientRect') # 'FP'
         extern clogical GetClientRect(chandle, cptr) user32
      endif
      if type('GetDC') # 'FP'
         extern chandle GetDC(chandle) user32
      endif
      if type('GetDeviceCaps') # 'FP'
         extern cint GetDeviceCaps(chandle, cint) gdi32
      endif
      if type('FillRect') # 'FP'
         extern clogical FillRect(chandle, cptr, chandle) user32
      endif
      if type('InvalidateRect') # 'FP'
         extern clogical InvalidateRect(chandle, cptr, clogical) user32
      endif
      if type('ReleaseDC') # 'FP'
         extern cint ReleaseDC(chandle, chandle) user32
      endif

      // get some environment settings to find out how 
      // many colors can be displayed

      hdc = GetDC(this.hwnd)
         this.BitsPerPixel = GetDeviceCaps(hdc, BITSPIXEL)
         this.NbrPlanes = GetDeviceCaps(hdc, PLANES)
         this.ColorBits = (this.BitsPerPixel * this.NbrPlanes)
      ReleaseDC(this.hwnd, hdc)

      // calculate how many steps to use in the fade based
      // on the number of colors that can be displayed

      do case
         case this.ColorBits = 32
            this.RgnCnt = 256
         case this.ColorBits = 24
            this.RgnCnt = 256
         case this.ColorBits = 16
            this.RgnCnt = 256
         case this.ColorBits = 15
            this.RgnCnt = 32
         case this.ColorBits = 8
            this.RgnCnt = 64
         case this.ColorBits = 4
            this.RgnCnt = 64
         otherwise
            this.RgnCnt = 64
            this.ColorBits = 4
      endcase

      return

   function onOpen
      // if we have a legal fade style
      if this.fadeStyle = 1 or this.fadeStyle = 2
         // fill our container
         this.anchor = 6
      endif
      return

   function Refresh
      // re-draws the gradient
      local r
      r = new RECT()
      GetClientRect(this.hwnd, r)
      InvalidateRect(this.hwnd, r, false)
      r.release()
      return


   function onPaint
      // this is where all the work is done

      local FillArea, AreaWidth, AreaHeight, ;
         RedStep, GreenStep, BlueStep, IntervalX, IntervalY, ;
         RedLevel, GreenLevel, BlueLevel, ps, hdc, hBrush

      // initialize our variables
      store 0 to FillArea, AreaWidth, AreaHeight, ;
         RedStep, GreenStep, BlueStep, IntervalX, IntervalY, ;
         RedLevel, GreenLevel, BlueLevel, ps, hdc, hBrush


      // first, get the size of the gradient
      FillArea := new RECT()
      GetClientRect(this.hwnd, FillArea)

      // store the height and width
      AreaHeight := FillArea.getHeight()
      AreaWidth := FillArea.getWidth()

      // calculate the increment of the steps for each color
      RedStep := (REDVAL(this.color1) - ;
         REDVAL(this.color2)) / this.RgnCnt
      GreenStep := (GREENVAL(this.color1) - ;
         GREENVAL(this.color2)) / this.RgnCnt
      BlueStep := (BLUEVAL(this.color1) - ;
         BLUEVAL(this.color2)) / this.RgnCnt

      // calculate the height & width of each step
      IntervalY := AreaHeight / this.RgnCnt
      IntervalX := AReaWidth / this.RgnCnt

      // initialize the three color values
      RedLevel := REDVAL(this.color2)
      GreenLevel := GREENVAL(this.color2)
      BlueLevel := BLUEVAL(this.color2)

      // throw away buffer for paintstructure
      ps := space(32)

      // get a device context on which to draw
      hdc := BeginPaint(this.hwnd, ps)

         // if we have a legal fadeStyle
         if this.fadeStyle == FADE_HORIZONTAL or ;
            this.fadeStyle == FADE_VERTICAL

            // process the area in steps
            for i = 0 to this.RgnCnt - 1

               // create a brush to paint with
               hBrush := CreateSolidBrush(RGB(;
                  RedLevel, GreenLevel, BlueLevel))

               if this.fadeStyle == FADE_VERTICAL
                  // decrement the height
                  AreaHeight -= IntervalY
                  // set the area to paint
                  FillArea.setTop(areaheight)
                  // draw a rect
                  FillRect(hdc, FillArea, hBrush)
                  // resize the rect
                  FillArea.setBottom(areaheight)
               else // FADE_HORIZONTAL
                  // decrement the width
                  AreaWidth -= IntervalX
                  // set the area to paint
                  FillArea.setLeft(Areawidth)
                  // draw a rect
                  FillRect(hdc, FillArea, hBrush)
                  // resize the rect
                  FillArea.setRight(areawidth)
               endif

               // delete the brush to prevent resource leakage
               DeleteObject(hBrush)

               // increment the color values
               RedLevel += RedStep
               GreenLevel += GreenStep
               BlueLevel += BlueStep

            endfor

            // now just finish off anything that might
            // not have been painted
            FillArea.setTop(0)
            FillArea.setLeft(0)
            hBrush := CreateSolidBrush(this.color1)
            FillRect(hdc, FillArea, hBrush)

            // delete the brush
            DeleteObject(hBrush)

         endif

      // done painting - release the handle
      EndPaint(this.hwnd, ps)

      // dispose of our RECT
      FillArea.release()

      // all done
      return

endclass

class RECT of struct

   this.string := space(8)

   function getLeft
      return int(this.getLong(0))

   function setLeft(n)
      return this.setLong(0, n)

   function getTop
      return int(this.getLong(4))

   function setTop(n)
      return this.setLong(4, n)

   function getRight
      return int(this.getLong(8))

   function setRight(n)
      return this.setLong(8, n)

   function getBottom
      return int(this.getLong(12))

   function setBottom(n)
      return this.setLong(12, n)

   function getWidth
      return int(this.getRight() - this.getLeft())

   function setWidth(n)
      return this.setRight(this.getLeft() + n)

   function getHeight
      return int(this.getBottom() - this.getTop())

   function setHeight(n)
      return this.setBottom(this.getTop() + n)

endclass

class struct of string

   function getLong(nIndex)
      local n
      n = this.getByte(nIndex) + ;
         bitlshift(this.getByte(nIndex + 1), 8) + ;
         bitlshift(this.getByte(nIndex + 2), 16) + ;
         bitlshift(this.getByte(nIndex + 3), 24)
      if bitset(n, 31)
         n := - bitnot(n)
      endif
      return n

   function setLong(nIndex, nValue)
      local bRet
      try
         bRet = true
         this.setByte(nIndex, bitand(nValue, 0xFF))
         this.setByte(nIndex + 1, bitrshift(bitand(nValue, 0xFF00), 8))
         this.setByte(nIndex + 2, bitrshift(bitand(nValue, 0xFF0000), 16))
         this.setByte(nIndex + 3, bitrshift(bitand(nValue, 0xFF000000), 24))
      catch(exception e)
         ? e.message, e.lineno
         bRet = false
      endtry
      return bRet

   function release
      this.string = ''
      release object this
      return

endclass
// eof: gradient.cc

