   /*
   --------------------------------------------------------
   Filename.....: winattr32.cc
   ClassName....: Winattr32
   Purpose......: Object to set another object's window properties.
   Programmer...: Rick Miller
                  Code used from work by:
                  David Wedwick and Romain Strieff in winattr.cc.
   Date.........: May 29, 2002
   Notes........: See below.
   Written for..: dB2K 0.1 and newer.
   Rev. History.: February 10, 2005  -  renamed all external functions.
                  November 30, 2004  -  full review with revisions.
   Dependencies.: None.
   Calls........: See x_InitExterns method.
   Called by....: Any
   Usage........: set procedure to :dUFLP:winattr32.cc additive
   Example......: See below and sample file.
   SampleFile...: winattr32.wfm.
   HelpFile.....: winattr32.hlp.
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      1) it is VERY IMPORTANT to call the
         close() or release() method
         when done using this object.
      2) use method calls after:
         form.open() or form.readmodal().
      3) changes for a form will take place when the
         form's windowstate property == 0 (Normal).
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
      // create:
      form.winattr32 = new Winattr32(form)

      // to set a property:
      form.winattr32.set("caption",     [<bool>])
      form.winattr32.set("changestate", [<bool>])
      form.winattr32.set("dlgframe",    [<bool>])
      form.winattr32.set("maximize",    [<bool>])
      form.winattr32.set("minimize",    [<bool>])
      form.winattr32.set("scroolh",     [<bool>])
      form.winattr32.set("scroolv",     [<bool>])
      form.winattr32.set("thickframe",  [<bool>])
      form.winattr32.set("sysmenu",     [<bool>])
      form.winattr32.set("window",      [<oref>])

      // destroy (very important that you close it,
      // or a circular reference to your form may
      // be created, and it won't be released ...):
      form.winattr32.close()
      form.winattr32 := null
      close procedure :dUFLP:winattr32.cc

   --------------------------------------------------------
   Properties:
   --------------------------------------------------------
      None are surfaced directly.

   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      close()
         calls to release
      get(<char> Keyword)
         returns current status.
      help([<char> Topic])
         opens the winattr32.hlp file when in the IDE.
      inspect()
         display properties in the inspector.
      release()
         release this object from memory.
      set(<char> Keyword, <pValu> value)
         change a property.
         returns true/false (success/fail)

   --------------------------------------------------------
   Methods compatible with winattr.cc for making a change.
   --------------------------------------------------------
   These methods return true/false (success/fail).
      border(     [<bool>])
      dlgframe(   [<bool>])
      maxbutton(  [<bool>])
      minbutton(  [<bool>])
      sysmenu(    [<bool>])
      thickframe( [<bool>])

   --------------------------------------------------------
   Keyword List used with the get and set methods:
   --------------------------------------------------------
      border, caption, changestate, dlgframe,
      maxbutton, maximize, minbutton, minimize,
      scrollh, scrollv, sysmenu, thickframe, window.
   */
//---------------------------------------------------------------//
//       Winattr32 constructor
//---------------------------------------------------------------//
#define  SWP_NOSIZE           0x0001
#define  SWP_NOMOVE           0x0002
#define  SWP_NOZORDER         0x0004
#define  SWP_NOACTIVATE       0x0010
#define  SWP_FRAMECHANGED     0x0020
#define  WINATTR32_VERSIONINFO   "5.0210"

class Winattr32(oWindow)
   // protected properties.
   protect changestate, runtime, window
   // protected methods.
   protect  x_ErrorMessage, x_GetBit, x_GetHwnd, x_InitExterns,;
            x_SetChangestate, x_SetWindow, x_SetWindowReference,;
            x_WindowTest

   class::x_InitExterns()
   this.changestate  =  false
   this.runtime   =  iif("run" $ lower(version(0)), true, false)
   this.versioninfo  =  WINATTR32_VERSIONINFO
   class::x_SetWindowReference(oWindow)

   //------------------------------------------------------------//
   //       end constructor   -  methods below
   //------------------------------------------------------------//
   function close
      this.release()
   return

   //------------------------------------------------------------//
   // Keywords:
   //    border, caption, changestate, dlgframe,
   //    maxbutton, maximize, minbutton, minimize,
   //    scrollh, scrollv, sysmenu, thickframe, window.
   function get(p1)
      Local bRet, cKey, cMess, nBit, nHwnd
      bRet  =  false
      cKey  =  lower("" + p1)
      cMess =  ""
      nBit  =  0
      nHwnd =  class::x_GetHwnd()
      if cKey == "changestate"
         bRet  := iif(this.changestate, true, false)
      elseif cKey == "window"
         bRet  := iif(nHwnd == 0, false, true)
      else
         nBit  := class::x_GetBit(cKey)
         cMess := "" + p1 + " is an Invalid parameter"
      endif
      if not nBit == 0
         // valid bit to check.
         if not nHwnd == 0
            // valid window reference.
            if (this.window.windowstate) == 1      // minimized.
               if nBit == 17 or nBit == 19
                  bRet := true
               endif
            elseif (this.window.windowstate) == 2  // maximized.
               if nBit == 16 or nBit == 19
                  bRet := true
               endif
            endif
            if not bRet    // normal windowstate.
               // get the status.
               bRet := bitSet(RMM_GetWindowLong(nHwnd, -16), nBit)
            endif
            cMess := ""
         else
           cMess := "hWnd is Invalid"
         endif
      endif
      if not empty(cMess)
         // unable to get.
         class::x_ErrorMessage("Unable to Get( " + p1 +;
            " ) in Winattr32" + chr(13) + chr(13) + cMess)
      endif
   return iif(bRet, true, false)
   // return true/false (status)

   //------------------------------------------------------------//
   function help(p1)
      if not this.runtime
         Local cFile, cHelp
         cFile =  left(program(1), rat(".", program(1)) - 1)
         cHelp =  setto("help")
         if file(cFile + ".hlp")
            set help to (cFile)
            if type("argVector(1)") == "C"
               private cTopc
               cTopc = "help " + p1
               &cTopc
               release cTopc
            else
               help intro
            endif
            set help to (cHelp)
         endif
      endif
   return

   //------------------------------------------------------------//
   function helpButton
      // from David Stone 02/19/2003
      // #define WS_EX_CONTEXTHELP htoi("00000400")
      // #define GWL_EXSTYLE -20
      // RMM_SetWindowLong(this.hwnd,GWL_EXSTYLE,WS_EX_CONTEXTHELP)
      Local bRet, nHwnd
      bRet  =  false
      nHwnd =  (this.oParent.hwnd)
      if this.maxbutton(false)
         if this.minbutton(false)
            RMM_SetWindowLong(nHwnd, -20, 1024)
            bRet  := true
         endif
      endif
   return iif(bRet, true, false)

   //------------------------------------------------------------//
   function inspect
      inspect(this)
   return

   //------------------------------------------------------------//
   function release
      this.window := null
      try
         release object this
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // Keywords:
   //    border, caption, changestate, dlgframe,
   //    maxbutton, maximize, minbutton, minimize,
   //    scrollh, scrollv, sysmenu, thickframe, window.
   function set(p1, p2)
      Local bRet, cKey, cMess, nArg
      bRet  =  false
      cKey  =  lower("" + p1)
      cMess =  "" + p1 + " is an Invalid parameter"
      nArg  =  argCount()
      if nArg == 0
      elseif nArg == 1
         cMess := "At least 2 parameters are required"
      elseif cKey == "changestate"
         bRet  := class::x_SetChangestate(p2)
         cMess := "" + p2 + " is an Invalid parameter"
      elseif cKey == "window"
         bRet  := class::x_SetWindowReference(p2)
      else
         if type("argVector(2)") == "L"
            // valid parameter type.
            Local nBit, nHwnd
            nBit  =  class::x_GetBit(cKey)
            if not nBit == 0
               // valid bit.
               nHwnd =  class::x_GetHwnd()
               if not nHwnd == 0
                  // valid hWnd.
                  cMess := ;
                     "Windowstate is " + iif( ;
                     (this.window.windowstate) == 2,;
                     "Maximized", "Minimized")
                  bBit  =  class::get(cKey)
                  bRet  := iif(p2 == bBit, true, false)
                  if not bRet
                     // a change has been requested.
                     // try to make change.
                     bRet  := class::x_SetWindow(nBit, nHwnd, bBit)
                  endif
               else
                  cMess += " or the hWnd is Invalid"
               endif
            endif
         else
            cMess := "2nd parameter -" + p2 + "- is Invalid"
         endif
      endif
      if not bRet
         // unable to set.
         class::x_ErrorMessage("Unable to Set( " + p1 + " ) " +;
               "in Winattr32" + chr(13) + chr(13) + cMess)
      endif
   return iif(bRet, true, false)
   // return true/false (success/fail)

   //------------------------------------------------------------//
   //       compatible methods below
   //       return true/false (success/fail)
   //------------------------------------------------------------//
   function border(p1)
   return class::set("Border", p1)
   //------------------------------------------------------------//
   function dlgframe(p1)
   return class::set("DlgFrame", p1)
   //------------------------------------------------------------//
   function maxbutton(p1)
   return class::set("MaxButton", p1)
   //------------------------------------------------------------//
   function minbutton(p1)
   return class::set("MinButton", p1)
   //------------------------------------------------------------//
   function sysmenu(p1)
   return class::set("SysMenu", p1)
   //------------------------------------------------------------//
   function thickframe(p1)
   return class::set("Thickframe", p1)
   //------------------------------------------------------------//
   //       end of compatible methods.
   //------------------------------------------------------------//

   //------------------------------------------------------------//
   // called by get, set methods.
   function x_GetBit(cKey)
      Local nRet
      nRet  =  0
      if cKey == "maxbutton"
         nRet := 16
      elseif cKey == "maximize"
         nRet := 16
      elseif cKey == "minbutton"
         nRet := 17
      elseif cKey == "minimize"
         nRet := 17
      elseif cKey == "thickframe"
         nRet := 18
      elseif cKey == "sysmenu"
         nRet := 19
      elseif cKey == "scrollh"
         nRet := 20
      elseif cKey == "scrollv"
         nRet := 21
      elseif cKey == "dlgframe"
         nRet := 22
      elseif cKey == "border"
         nRet := 23
      elseif cKey == "caption"
         nRet := 23
      endif
   return int(nRet)
   // return 0 or bitPosition

   //------------------------------------------------------------//
   function x_ErrorMessage(cMess)
      if not this.runtime
         // in the IDE.
         msgbox(cMess, "Error in winattr32.cc", 48)
      endif
   return

   //------------------------------------------------------------//
   // called by get, set methods.
   function x_GetHwnd
      Local nRet
      nRet  =  0
      if class::x_WindowTest()
         nRet  := 0 + this.window.hwnd
      endif
   return int(nRet)
   // return 0 or hWnd.

   //------------------------------------------------------------//
   // initialize external functions.
   function x_InitExterns
      if not type("RMM_GetWindowLong") == "FP"
         extern CLONG RMM_GetWindowLong(CHANDLE, CINT) ;
            user32 from "GetWindowLongA"
      endif
      if not type("RMM_SetWindowLong") == "FP"
         extern CLONG RMM_SetWindowLong(CHANDLE, CINT, CLONG) ;
            user32 from "SetWindowLongA"
      endif
      if not type("RMM_SetWindowPos") == "FP"
         extern CLOGICAL RMM_SetWindowPos( ;
            CHANDLE, CHANDLE, CINT, CINT, CINT, CINT, CUINT) ;
            user32 from "SetWindowPos"
      endif
   return

   //------------------------------------------------------------//
   function x_SetChangestate(p1)
      if type("argVector(1)") == "L"
         this.changestate  := iif(p1, true, false)
      endif
   return type("argVector(1)") == "L"

   //------------------------------------------------------------//
   function x_SetWindow(nBit, nHwnd, bStat)
      Local bRet
      bRet  =  false
      if not this.window.windowstate == 0
         if this.changestate
            // change windowstate to 0.
            this.window.windowstate := 0
         endif
      endif
      if this.window.windowstate == 0
         // changes are made only when the windowstate == 0.
         Local nMask, nStyl
         // get the current window style.
         nStyl =  RMM_GetWindowLong(nHwnd, -16)
         // get the change amount.
         nMask =  bitlshift(1, nBit)
         if bStat
            nStyl -= nMask
         else
            nStyl += nMask
         endif
         // change the window.
         RMM_SetWindowLong(nHwnd, -16, nStyl)
         // display the changes.
         bRet := ;
            RMM_SetWindowPos(nHwnd, 0, 0, 0, 0, 0,;
            SWP_FRAMECHANGED + SWP_NOZORDER +;
            SWP_NOACTIVATE + SWP_NOSIZE + SWP_NOMOVE)
      endif
   return iif(bRet, true, false)
   // return true/false (success/fail)

   //------------------------------------------------------------//
   function x_SetWindowReference
      parameters p1
      Local bRet
      bRet  =  type("p1.hwnd") == "N"
      if bRet
         this.window =  p1
      endif
   return iif(bRet, true, false)

   //------------------------------------------------------------//
   // remove any object reference that is not valid.
   function x_WindowTest
      Local bRet
      bRet  =  type("this.window.windowstate") == "N"
      if not bRet
         this.window := null
      endif
   return iif(bRet, true, false)
   // return true/false (valid window object)

endclass
//---------------------------------------------------------------//
//          end of Winattr32.
//---------------------------------------------------------------//

